from __future__ import annotations

import asyncio
import contextlib
import functools
import json
import os
import re
import select
import sys
import uuid
from enum import Enum
from pathlib import Path
from typing import Optional, Union

import boto3
import click
import questionary
from aws_error_utils.aws_error_utils import errors
from botocore.exceptions import ClientError, NoCredentialsError
from pydantic.json import pydantic_encoder
from pydantic.types import SecretStr

from iambic.config.dynamic_config import (
    CURRENT_IAMBIC_VERSION,
    Config,
    ExtendsConfig,
    ExtendsConfigKey,
    load_config,
    process_config,
)
from iambic.config.utils import (
    check_and_update_resource_limit,
    resolve_config_template_path,
)
from iambic.core.context import ctx
from iambic.core.iambic_enum import Command, IambicManaged
from iambic.core.logger import log
from iambic.core.models import ExecutionMessage
from iambic.core.parser import load_templates
from iambic.core.template_generation import get_existing_template_map
from iambic.core.utils import gather_templates, yaml
from iambic.github.utils import create_workflow_files
from iambic.plugins.v0_1_0.aws.cloud_formation.utils import (
    create_iambic_eventbridge_stacks,
    create_iambic_role_stacks,
    create_spoke_role_stack,
)
from iambic.plugins.v0_1_0.aws.handlers import apply as aws_apply
from iambic.plugins.v0_1_0.aws.handlers import import_aws_resources
from iambic.plugins.v0_1_0.aws.handlers import load as aws_load
from iambic.plugins.v0_1_0.aws.iam.role.models import (
    AWS_IAM_ROLE_TEMPLATE_TYPE,
    AwsIamRoleTemplate,
)
from iambic.plugins.v0_1_0.aws.iambic_plugin import AWSConfig
from iambic.plugins.v0_1_0.aws.models import (
    ARN_RE,
    IAMBIC_SPOKE_ROLE_NAME,
    AWSAccount,
    AWSIdentityCenter,
    AWSOrganization,
    BaseAWSOrgRule,
    Partition,
    get_hub_role_arn,
    get_spoke_role_arn,
)
from iambic.plugins.v0_1_0.aws.utils import (
    RegionName,
    get_identity_arn,
    is_valid_account_id,
)
from iambic.plugins.v0_1_0.azure_ad.handlers import import_azure_ad_resources
from iambic.plugins.v0_1_0.azure_ad.iambic_plugin import AzureADConfig
from iambic.plugins.v0_1_0.azure_ad.models import AzureADOrganization
from iambic.plugins.v0_1_0.google_workspace.handlers import import_google_resources
from iambic.plugins.v0_1_0.google_workspace.iambic_plugin import (
    GoogleProject,
    GoogleSubject,
    GoogleWorkspaceConfig,
)
from iambic.plugins.v0_1_0.okta.handlers import import_okta_resources
from iambic.plugins.v0_1_0.okta.iambic_plugin import OktaConfig, OktaOrganization

AWS_SECRETS = "AWS Secrets"
LOCALLY = "Locally"

CUSTOM_AUTO_COMPLETE_STYLE = questionary.Style(
    [
        ("answer", "fg:#0A886A"),
        ("selected", "bold bg:#000000"),
    ]
)


class Operation(Enum):
    ADDED = "added"
    UPDATED = "updated"
    DELETED = "deleted"


def clear_stdin_buffer():
    """Clears the standard input (stdin) buffer.

    This function reads and discards any input that may be present in the standard
    input (stdin) buffer. This can be useful in cases where previous input may be
    interfering with the desired behavior of a subsequent input operation.

    Args:
        None

    Returns:
        None
    """
    # Warning: This appears to work fine in a real terminal,
    # but not VSCode's Debug Terminal

    r, _, _ = select.select([sys.stdin], [], [], 0)
    while r:
        # If there is input waiting, read and discard it.
        sys.stdin.readline()
        r, _, _ = select.select([sys.stdin], [], [], 0)


def monkeypatch_questionary():
    """Monkeypatches the questionary functions in use to clear stdin buffer."""
    original_functions = {
        "prompt": questionary.prompt,
        "ask": questionary.Question.ask,
        "unsafe_ask": questionary.Question.unsafe_ask,
        "select": questionary.select,
    }

    def patched_function(original_function):
        @functools.wraps(original_function)
        def wrapper(*args, **kwargs):
            clear_stdin_buffer()
            return original_function(*args, **kwargs)

        return wrapper

    for function_name, original_function in original_functions.items():
        setattr(questionary, function_name, patched_function(original_function))


def get_secret_dict_with_val(pydantic_model, **kwargs) -> dict:
    def show_secrets_encoder(obj):
        if isinstance(obj, SecretStr):
            return obj.get_secret_value()
        else:
            return pydantic_encoder(obj)

    return json.loads(pydantic_model.json(encoder=show_secrets_encoder, **kwargs))


def set_aws_region(question_text: str, default_val: Union[str, RegionName]) -> str:
    default_val = default_val if isinstance(default_val, str) else default_val.value
    choices = [default_val] + [e.value for e in RegionName if e.value != default_val]
    return questionary.select(
        question_text, choices=choices, default=default_val
    ).unsafe_ask()


def set_aws_account_partition(default_val: Union[str, Partition]) -> str:
    return questionary.select(
        "Which AWS partition is the account on?",
        choices=[e.value for e in Partition],
        default=default_val if isinstance(default_val, str) else default_val.value,
    ).unsafe_ask()


def set_aws_role_arn(account_id: str):
    while True:
        click.echo(
            "\n(Optional) Provide a role arn that CloudFormation will assume to create the stack(s) "
            "or hit enter to use your current access."
        )
        role_arn = questionary.text("(Optional) Role ARN").unsafe_ask()
        if not role_arn or (account_id in role_arn and re.search(ARN_RE, role_arn)):
            return role_arn or None
        else:
            log.warning(
                "The role ARN must be a valid ARN for the account you are configuring.",
                expected_account_id=account_id,
                provided_role_arn=role_arn,
            )


def set_aws_is_read_only() -> bool:
    click.echo(
        "\nGrant IambicSpokeRole write access to IAM and IdentityCenter?\n"
        "If set to 'no', this will limit IAMbic's capabilities to import-only."
    )
    return not bool(
        questionary.confirm(
            "Grant IambicSpokeRole write access?",
            default=True,
        ).unsafe_ask()
    )


def set_required_text_value(human_readable_name: str, default_val: str = None):
    while True:
        if response := questionary.text(
            human_readable_name,
            default=default_val or "",
        ).unsafe_ask():
            return response
        else:
            print("Please enter a valid response.")


def set_required_secure_text_value(
    human_readable_name: str,
    default_val: Optional[SecretStr] = None,
):
    while True:
        if response := questionary.text(
            human_readable_name,
            default=SecretStr(default_val or "").get_secret_value(),
        ).unsafe_ask():
            return SecretStr(response.replace("\\n", "\n"))
        else:
            print("Please enter a valid response.")


def set_idp_name(default_val: str = None):
    return set_required_text_value("What is the Identity Provider Name?", default_val)


def set_tenant_id(default_val: str = None):
    return set_required_text_value("What is the Tenant ID?", default_val)


def set_client_id(default_val: str = None):
    return set_required_text_value("What is the Client ID?", default_val)


def set_client_secret(default_val: str = None):
    return set_required_text_value("What is the Client Secret?", default_val)


def set_okta_org_url(default_val: str = None):
    return set_required_text_value("What is the Organization URL?", default_val)


def set_okta_api_token(default_val: str = None):
    return set_required_text_value("What is the Okta API Token?", default_val)


def set_google_subject(default_domain: str = None, default_service: str = None) -> dict:
    return {
        "domain": set_required_text_value("What is the Google Domain?", default_domain),
        "service_account": set_required_text_value(
            "What is the Google Service Account?", default_service
        ),
    }


def set_google_project_type(default_val: str = None):
    return set_required_text_value(
        "What is the Google Project Type?", default_val or "service_account"
    )


def set_google_project_id(default_val: str = None):
    return set_required_text_value("What is the Project ID?", default_val)


def set_google_private_key(default_val: SecretStr = None):
    return set_required_secure_text_value("What is the Private Key?", default_val)


def set_google_private_key_id(default_val: str = None):
    return set_required_text_value("What is the Private Key ID?", default_val)


def set_google_client_email(default_val: str = None):
    return set_required_text_value("What is the Client E-Mail?", default_val)


def set_google_auth_uri(default_val: str = None):
    return set_required_text_value("What is the Auth URI?", default_val)


def set_google_token_uri(default_val: str = None):
    return set_required_text_value("What is the Token URI?", default_val)


def set_google_auth_provider_cert_url(default_val: str = None):
    return set_required_text_value(
        "What is the auth_provider_x509_cert_url?", default_val
    )


def set_google_client_cert_url(default_val: str = None):
    return set_required_text_value("What is the client_x509_cert_url?", default_val)


def set_identity_center(
    region: str = RegionName.us_east_1,
) -> AWSIdentityCenter:
    identity_center = AWSIdentityCenter()
    identity_center.region = set_aws_region(
        "What region is your Identity Center (SSO) set to?", region
    )
    return identity_center


def confirm_command_exe(
    provider_type: str, operation: Operation, requires_sync: bool = False
):
    command_type = "import"
    if requires_sync:
        command_type = (
            "an apply of applicable templates that are NOT in the account and an import"
        )

    if operation == Operation.ADDED:
        operation_str = f"{operation.value} to"
    elif operation == Operation.UPDATED:
        operation_str = f"{operation.value} in"
    elif operation == Operation.DELETED:
        operation_str = f"{operation.value} from"
    else:
        raise ValueError(f"Invalid operation: {operation}")

    click.echo(
        f"\nTo preserve these changes, {command_type} must be run to sync your templates."
    )
    if not questionary.confirm("Proceed?").unsafe_ask():
        click.echo(
            f"\nThe {provider_type} will not be {operation_str} the config and wizard will exit."
        )
        if questionary.confirm("Proceed?").unsafe_ask():
            log.info("Exiting...")
            sys.exit(0)


class ConfigurationWizard:
    def __init__(self, repo_dir: str):
        # TODO: Handle the case where the config file exists but is not valid
        self.existing_role_template_map = {}
        self.aws_account_map = {}
        self.repo_dir = repo_dir
        self._has_cf_permissions = None
        self._cf_role_arn = None
        self._assume_as_arn = None
        self.caller_identity = {}
        self.profile_name = ""
        self._default_region = None

        asyncio.run(self.set_config_details())
        check_and_update_resource_limit(self.config)
        log.debug("Starting configuration wizard", config_path=self.config_path)

    @property
    def has_cf_permissions(self):
        if self._has_cf_permissions is None:
            try:
                click.echo(
                    f"\nThis requires that you have the ability to "
                    f"create CloudFormation stacks, stack sets, and stack set instances.\n"
                    f"If you are using an AWS Organization, be sure that trusted access is enabled.\n"
                    f"You can check this using the AWS Console:\n  "
                    f"https://{self.aws_default_region}.console.aws.amazon.com/organizations/v2/home/services/CloudFormation%20StackSets"
                )
                self._has_cf_permissions = questionary.confirm("Proceed?").unsafe_ask()
            except KeyboardInterrupt:
                log.info("Exiting...")
                sys.exit(0)

        return self._has_cf_permissions

    @property
    def assume_as_arn(self):
        if self._assume_as_arn is None:
            current_arn = get_identity_arn(self.caller_identity)
            click.echo(
                "\nProvide the ARN of the identity (e.g. IAM User, Group, or Role) "
                "that will be able to access the hub role.\n"
                "Note: Access to this identity is required to use IAMbic locally.",
            )
            self._assume_as_arn = questionary.text(
                "Identity ARN",
                default=current_arn,
            ).ask()

        return self._assume_as_arn

    @property
    def cf_role_arn(self):
        if self._cf_role_arn is None:
            self._cf_role_arn = set_aws_role_arn(self.hub_account_id)

        return self._cf_role_arn

    @property
    def has_aws_account_or_organizations(self) -> bool:
        return self.config.aws and (
            self.config.aws.accounts or self.config.aws.organizations
        )

    @property
    def aws_default_region(self) -> str:
        default_val = RegionName.us_east_1.value

        if self._default_region:
            return self._default_region
        elif self.has_aws_account_or_organizations:
            if self.config.aws.organizations:
                self._default_region = self.config.aws.organizations[0].region_name
            else:
                hub_account = [
                    account
                    for account in self.config.aws.accounts
                    if account.hub_role_arn
                ]
                if hub_account:
                    self._default_region = hub_account[0].region_name
                else:
                    self._default_region = set_aws_region(
                        "What region should IAMbic use?", default_val
                    )
        else:
            self._default_region = set_aws_region(
                "What region should IAMbic use?", default_val
            )

        return self._default_region

    @property
    def secrets_message(self) -> tuple[str, bool]:
        """Returns if secrets are configured"""
        secret_in_config = bool(self.config.extends)

        if (
            secret_in_config
            and self.config.extends[0].key == ExtendsConfigKey.AWS_SECRETS_MANAGER
        ):
            secret_question_text = (
                "This requires the ability to update the AWS Secrets Manager secret."
            )
        elif (
            secret_in_config
            and self.config.extends[0].key == ExtendsConfigKey.LOCAL_FILE
        ):
            secret_question_text = (
                "This requires permissions to update secrets in your secrets file."
            )
        else:
            secret_question_text = "Secrets must be created."

        return (secret_question_text, secret_in_config)

    @property
    def has_okta_configured(self) -> bool:
        return self.config.okta and self.config.okta.organizations

    def secret_config_dir(self, secrets_file: str = "secrets.yaml") -> str:
        return f"{Path(self.config_path).parent.joinpath(secrets_file)}"

    async def set_config_details(self):
        try:
            self.config_path = str((await resolve_config_template_path(self.repo_dir)))
        except RuntimeError:
            self.config_path = f"{self.repo_dir}/iambic_config.yaml"

        if os.path.exists(self.config_path) and os.path.getsize(self.config_path) != 0:
            log.info("Found existing configuration file", config_path=self.config_path)
            try:
                self.config = await load_config(self.config_path)
            except Exception as err:
                log.error(
                    "Unable to load existing configuration file",
                    config_path=self.config_path,
                    error=repr(err),
                )
                sys.exit(1)
        else:
            # Create a stubbed out config file to use for the wizard
            self.config_path = f"{self.repo_dir}/iambic_config.yaml"
            base_config: Config = Config(
                file_path=self.config_path, version=CURRENT_IAMBIC_VERSION
            )

            self.config = await process_config(
                base_config, self.config_path, base_config.dict()
            )

    def resolve_aws_profile_defaults_from_env(self) -> str:
        if "AWS_ACCESS_KEY_ID" in os.environ:
            # Environment variables has 1st priority
            profile_name = "None"
            log.info("Using AWS credentials from environment", profile=profile_name)
        elif profile_name := os.environ.get("AWS_PROFILE"):
            # Explicit profile has 2nd priority
            log.info("Using AWS profile from environment", profile=profile_name)
        elif profile_name := os.environ.get("AWS_DEFAULT_PROFILE"):
            # Fallback profile has 3rd priority
            log.info("Using AWS default profile from environment", profile=profile_name)
        else:
            # User has to direct the wizard at this point since
            # we cannot reason what credential to use.
            profile_name = "None"
            log.info(
                "Not able detect a standard credential provider chain",
                profile=profile_name,
            )

        return profile_name

    def set_aws_profile_name(
        self, question_text: str = None, allow_none: bool = False
    ) -> Union[str, None]:
        questionary_params = {}
        boto3_session = boto3.Session(region_name=self.aws_default_region)
        available_profiles = boto3_session.available_profiles
        if allow_none:
            available_profiles.insert(0, "None")

        default_profile = self.resolve_aws_profile_defaults_from_env()
        if default_profile != "None":
            questionary_params["default"] = default_profile
            available_profiles.append(default_profile)

        if not question_text:
            click.echo(
                f"\nWe couldn't find your AWS credentials, or they're not linked to the Hub Account ({self.hub_account_id}). "
                "The specified AWS credentials need to be able to create CloudFormation stacks, stack sets, and stack set instances. "
                "Please provide an AWS profile to use for this operation or restart the wizard with valid AWS credentials."
            )
            question_text = "AWS Profile"

        try:
            if len(available_profiles) == 0:
                log.error(
                    "Please create an AWS profile with access to the Hub Account. "
                    "See https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-profiles.html"
                )
                sys.exit(0)
            elif len(available_profiles) < 10:
                profile_name = questionary.select(
                    question_text, choices=available_profiles, **questionary_params
                ).unsafe_ask()
            else:
                profile_name = questionary.autocomplete(
                    question_text,
                    choices=available_profiles,
                    style=CUSTOM_AUTO_COMPLETE_STYLE,
                    **questionary_params,
                ).unsafe_ask()
        except KeyboardInterrupt:
            log.info("Exiting...")
            sys.exit(0)

        return profile_name if profile_name != "None" else None

    def set_boto3_session(self, allow_none=False):
        self._has_cf_permissions = True
        while True:
            try:
                profile_name = self.set_aws_profile_name(allow_none=allow_none)
                self.boto3_session = boto3.Session(
                    profile_name=profile_name, region_name=self.aws_default_region
                )
                self.caller_identity = self.boto3_session.client(
                    "sts", region_name=self.aws_default_region
                ).get_caller_identity()
                selected_hub_account_id = self.caller_identity.get("Arn").split(":")[4]
                if selected_hub_account_id != self.hub_account_id:
                    log.error(
                        "The selected profile does not have access to the Hub Account. Please try again.",
                        required_account_id=self.hub_account_id,
                        selected_account_id=selected_hub_account_id,
                    )
                    continue
            except errors.FileNotFoundError as err:
                log.error(
                    (
                        "We are unable to authenticate to AWS because the provided profile name has a credential_process "
                        "rule setup in your AWS configuration file. Either remove the credential_process rule or use the "
                        "AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY environment variables."
                    ),
                    error=str(err),
                )
                continue
            except errors.ClientError as err:
                log.info(
                    "Unable to create a session for the provided profile name. Please try again.",
                    error=str(err),
                )
                continue
            except errors.ProfileNotFound as err:
                log.info(
                    "Selected profile doesn't exist. Please try again.",
                    error=str(err),
                )
                continue
            except errors.InvalidClientTokenId:
                log.error(
                    "AWS returned an error indicating that the provided credentials are invalid. Somethings to try:"
                    "\n - Ensure that the credentials are correct"
                    "\n - Ensure that the credentials are for the correct AWS account"
                    "\n - Ensure that the credentials have the correct permissions"
                    "\n - Ensure that the credentials are not expired"
                    "\n - Ensure that the credentials are not for a federated user"
                )
                continue

            self.profile_name = profile_name
            with contextlib.suppress(
                ClientError, NoCredentialsError, FileNotFoundError
            ):
                self.autodetected_org_settings = self.boto3_session.client(
                    "organizations", region_name=self.aws_default_region
                ).describe_organization()["Organization"]
            break

    def get_boto3_session_for_account(self, account_id: str, region_name: str = None):
        # This need to follow standard credentials provider chain
        if not region_name:
            region_name = self.aws_default_region

        if account_id == self.hub_account_id:
            if not self.profile_name:
                if os.getenv("AWS_ACCESS_KEY_ID"):
                    # environment variables credentials detected
                    self.profile_name = None
                elif profile_name := os.getenv("AWS_PROFILE"):
                    self.profile_name = profile_name
                else:
                    self.profile_name = self.set_aws_profile_name(
                        "Please specify the profile to use to access to the AWS Account.",
                        allow_none=False,
                    )
            return self.boto3_session, self.profile_name
        else:
            profile_name = self.set_aws_profile_name(
                "Please specify the profile to use to access to the AWS Account.\n"
                "If None is selected the AWS Account will be skipped.",
                allow_none=True,
            )
            if not profile_name:
                log.info("Unable to add the AWS Account without a session.")
                return None, None
            return (
                boto3.Session(profile_name=profile_name, region_name=region_name),
                profile_name,
            )

    async def run_import_aws_resources(self):
        log.info("Importing AWS identities")
        current_command = ctx.command
        ctx.command = Command.IMPORT

        exe_message = ExecutionMessage(
            execution_id=str(uuid.uuid4()),
            command=Command.IMPORT,
            provider_type="aws",
        )
        await import_aws_resources(
            exe_message,
            self.config.aws,
            self.repo_dir,
        )

        ctx.command = current_command

    async def sync_config_aws_accounts(self, accounts: list[AWSAccount]):
        if not (
            len(self.config.aws.accounts)
            > self.config.aws.min_accounts_required_for_wildcard_included_accounts
        ):
            await self.run_import_aws_resources()
            return

        templates = await gather_templates(self.repo_dir, "AWS.*")
        if templates:
            current_command = ctx.command
            ctx.command = Command.CONFIG_DISCOVERY
            log.info(
                "Applying templates to provision identities to the account(s). "
                "This will NOT overwrite any resources that already exist on the account(s). ",
                accounts=[account.account_name for account in accounts],
            )
            exe_message = ExecutionMessage(
                execution_id=str(uuid.uuid4()),
                command=Command.APPLY,
                provider_type="aws",
            )
            sub_config = self.config.aws.copy()
            sub_config.accounts = accounts
            await aws_apply(exe_message, sub_config, load_templates(templates))
            ctx.command = current_command

        await self.run_import_aws_resources()

    async def sync_config_aws_org(self, run_config_discovery: bool = True):
        if not self.config.aws:
            self.config.aws = AWSConfig()

        self.aws_account_map = {}
        current_command = ctx.command

        try:
            if run_config_discovery:
                exe_message = ExecutionMessage(
                    execution_id=str(uuid.uuid4()),
                    command=Command.CONFIG_DISCOVERY,
                )
                await self.config.run_discover_upstream_config_changes(
                    exe_message, self.repo_dir
                )
            await self.config.aws.set_identity_center_details()
        except Exception as err:
            log.info("Failed to refresh AWS accounts", error=err)

        ctx.command = current_command
        self.aws_account_map = {
            account.account_id: account for account in self.config.aws.accounts
        }

    async def run_import_google_workspace_resources(self):
        log.info("Importing Google Workspace identities")
        current_command = ctx.command
        ctx.command = Command.IMPORT

        exe_message = ExecutionMessage(
            execution_id=str(uuid.uuid4()),
            command=Command.IMPORT,
            provider_type="google_workspace",
        )
        await import_google_resources(
            exe_message,
            self.config.google_workspace,
            self.repo_dir,
        )

        ctx.command = current_command

    async def run_import_azure_ad_resources(self):
        log.info("Importing Azure AD identities")
        current_command = ctx.command
        ctx.command = Command.IMPORT

        exe_message = ExecutionMessage(
            execution_id=str(uuid.uuid4()),
            command=Command.IMPORT,
            provider_type="azure_ad",
        )
        await import_azure_ad_resources(
            exe_message,
            self.config.azure_ad,
            self.repo_dir,
        )

        ctx.command = current_command

    async def run_import_okta_resources(self):
        log.info("Importing Okta identities")
        current_command = ctx.command
        ctx.command = Command.IMPORT

        exe_message = ExecutionMessage(
            execution_id=str(uuid.uuid4()),
            command=Command.IMPORT,
            provider_type="okta",
        )
        await import_okta_resources(
            exe_message,
            self.config.okta,
            self.repo_dir,
        )

        ctx.command = current_command

    async def save_and_deploy_changes(self, role_template: AwsIamRoleTemplate):
        log.info(
            "Writing changes locally and deploying updates to AWS",
            role_name=role_template.properties.role_name,
        )

        self.config.write()
        role_template.write(exclude_unset=False)
        await role_template.apply(self.config.aws)

    def configuration_wizard_aws_account_add(self):  # noqa: C901
        if not self.has_cf_permissions:
            log.info(
                "Unable to edit this attribute without CloudFormation permissions."
            )
            return

        is_hub_account = bool(
            not self.config.aws.accounts and not self.config.aws.organizations
        )
        requires_sync = bool(
            not is_hub_account
            and (
                len(self.config.aws.accounts)
                > self.config.aws.min_accounts_required_for_wildcard_included_accounts
            )
        )

        if is_hub_account:
            account_id = self.hub_account_id
            account_name = set_required_text_value(
                "What is the name of the AWS Account?"
            )
            default_region = self.aws_default_region
            self.config.aws.spoke_role_is_read_only = set_aws_is_read_only()
            click.echo(
                "\nIAMbic requires Hub and Spoke roles to be created which is deployed using CloudFormation.\n"
                "To review the templates used or deploy them manually, the templates used can be found here:\n"
                "https://github.com/noqdev/iambic/tree/main/iambic/plugins/v0_1_0/aws/cloud_formation/templates\n"
                "If you have already manually deployed the templates, answer yes to proceed.\n"
                "IAMbic will validate that your stacks have been deployed successfully and will not attempt to replace them."
            )

            if not questionary.confirm("Proceed?").unsafe_ask():
                log.info(
                    "Unable to add the AWS Account without creating the required roles."
                )
                return

        else:
            if requires_sync:
                click.echo(
                    "\nAdding this account will require a sync to be run.\n"
                    "This is to apply any matching templates to the account if the resource does not already exist.\n"
                    "Then, the account resources will be imported into Iambic."
                )
                if not questionary.confirm("Proceed?").unsafe_ask():
                    log.info(
                        "Unable to add the AWS account without creating the required role."
                    )
                    return

            account_id = questionary.text(
                "What is the AWS Account ID? Example: `12345689012`"
            ).unsafe_ask()
            account_name = questionary.text(
                "What is the name of the AWS Account?"
            ).unsafe_ask()
            if not is_valid_account_id(account_id):
                log.info("Invalid AWS Account ID")
                return
            elif account_id in list(self.aws_account_map.keys()):
                log.info("AWS Account already exists in the configuration")
                return

            default_region = set_aws_region(
                "What region should IAMbic use?", self.aws_default_region
            )
            role_arn = set_aws_role_arn(account_id)
            assume_as_arn = self.assume_as_arn

            click.echo(
                "\nIAMbic requires Hub and Spoke roles to be created which is deployed using CloudFormation.\n"
                "To review the templates used or deploy them manually, the template used can be found here:\n"
                "https://github.com/noqdev/iambic/blob/main/iambic/plugins/v0_1_0/aws/cloud_formation/templates/IambicSpokeRole.yml\n"
                "If you have already manually deployed the templates, answer yes to proceed.\n"
                "IAMbic will validate that your stacks have been deployed successfully and will not attempt to replace them."
            )

            if not questionary.confirm("Proceed?").unsafe_ask():
                log.info(
                    "Unable to add the AWS account without creating the required role."
                )
                return

        read_only = self.config.aws.spoke_role_is_read_only
        session, profile_name = self.get_boto3_session_for_account(
            account_id, region_name=default_region
        )
        if not session:
            return

        if is_hub_account and not profile_name:
            profile_name = self.profile_name
        elif not is_hub_account:
            profile_name = None

        cf_client = session.client("cloudformation", region_name=default_region)

        if is_hub_account:
            created_successfully = asyncio.run(
                create_iambic_role_stacks(
                    cf_client=cf_client,
                    hub_account_id=account_id,
                    assume_as_arn=assume_as_arn,
                    role_arn=role_arn,
                    read_only=read_only,
                )
            )
            if not created_successfully:
                log.error("Failed to create the required IAMbic roles. Exiting.")
                sys.exit(0)
        else:
            created_successfully = asyncio.run(
                create_spoke_role_stack(
                    cf_client=cf_client,
                    hub_account_id=account_id,
                    role_arn=role_arn,
                    read_only=read_only,
                )
            )
            if not created_successfully:
                log.error(
                    "Failed to create the required IAMbic role. Account not added."
                )
                return

        iambic_managed = (
            IambicManaged.IMPORT_ONLY if read_only else IambicManaged.READ_AND_WRITE
        )
        account = AWSAccount(
            account_id=account_id,
            account_name=account_name,
            spoke_role_arn=get_spoke_role_arn(account_id),
            iambic_managed=iambic_managed,
            aws_profile=profile_name,
            default_region=default_region,
        )
        if is_hub_account:
            account.hub_role_arn = get_hub_role_arn(account_id)
        # account.partition = set_aws_account_partition(account.partition)

        confirm_command_exe("AWS Account", Operation.ADDED, requires_sync=requires_sync)

        self.config.aws.accounts.append(account)

        if is_hub_account:
            check_and_update_resource_limit(self.config)
            asyncio.run(self.run_import_aws_resources())
        else:
            self.config.aws = asyncio.run(aws_load(self.config.aws))
            asyncio.run(self.sync_config_aws_accounts([account]))

    def configuration_wizard_aws_account_edit(self):
        account_names = [
            account.account_name
            for account in self.config.aws.accounts
            if not account.org_id
        ]
        account_id_to_config_elem_map = {
            account.account_id: elem
            for elem, account in enumerate(self.config.aws.accounts)
        }
        if len(account_names) > 1:
            action = questionary.autocomplete(
                "Which AWS Account would you like to edit?",
                choices=["Go back", *account_names],
                style=CUSTOM_AUTO_COMPLETE_STYLE,
            ).unsafe_ask()
            if action == "Go back":
                return
            account = next(
                (
                    account
                    for account in self.config.aws.accounts
                    if account.account_name == action
                ),
                None,
            )
            if not account:
                log.debug("Could not find AWS Account")
                return
        elif not account_names:
            log.info(
                "No editable accounts found.\n"
                "TIP: An AWS account cannot be edited if it attached to an Organization in the config."
            )
            return
        else:
            account = self.config.aws.accounts[0]

        choices = ["Go back", "Update IAMbic control", "Update Region"]
        if not account.org_id:
            choices.append("Update name")

        while True:
            action = questionary.select(
                "What would you like to do?",
                choices=choices,
            ).unsafe_ask()
            if action == "Go back":
                return
            elif action == "Update name":
                account.default_region = set_aws_region(
                    "What region should IAMbic use?", account.default_region
                )
            elif action == "Update name":
                account.account_name = questionary.text(
                    "What is the name of the AWS Account?",
                    default=account.account_name,
                ).unsafe_ask()

            confirm_command_exe("AWS Account", Operation.UPDATED)

            self.config.aws.accounts[
                account_id_to_config_elem_map[account.account_id]
            ] = account
            asyncio.run(self.run_import_aws_resources())
            self.config.write()

    def setup_aws_configuration(self):
        if self.has_aws_account_or_organizations:
            self.hub_account_id = self.config.aws.hub_role_arn.split(":")[4]
            self.spoke_role_is_read_only = self.config.aws.spoke_role_is_read_only
        else:
            self.hub_account_id = None

        try:
            if getattr(self, "boto3_session", None) is None:
                # need bootstrapping
                self.boto3_session = boto3.Session(region_name=self.aws_default_region)
                self.autodetected_org_settings = {}

            default_caller_identity = self.boto3_session.client(
                "sts", region_name=self.aws_default_region
            ).get_caller_identity()
            caller_arn = get_identity_arn(default_caller_identity)
            default_hub_account_id = caller_arn.split(":")[4]
        except (
            AttributeError,
            IndexError,
            NoCredentialsError,
            ClientError,
            FileNotFoundError,
        ):
            default_hub_account_id = None
            default_caller_identity = {}

        if not self.hub_account_id:
            while True:
                click.echo(
                    "\nTo get started with the IAMbic setup wizard, you'll need an AWS account.\n"
                    "This is where IAMbic will deploy its main role.\n"
                    "If you have an AWS Organization, that account will be your hub account.\n"
                    "Review to-be-created IAMbic roles at\n"
                    "https://docs.iambic.org/reference/aws_hub_and_spoke_roles"
                )
                self.hub_account_id = set_required_text_value(
                    "Which Account ID should we use to deploy the IAMbic hub role?",
                    default_val=default_hub_account_id,
                )
                if is_valid_account_id(self.hub_account_id):
                    break

        if self.hub_account_id == default_hub_account_id:
            identity_arn = get_identity_arn(default_caller_identity)
            click.echo(
                f"\nIAMbic detected you are using {identity_arn} for AWS access.\n"
                f"This identity will require the ability to create"
                f"CloudFormation stacks, stack sets, and stack set instances."
            )
            if questionary.confirm("Would you like to use this identity?").ask():
                self.caller_identity = default_caller_identity
                # If we are going to use the default_caller_identity,
                # we need to set teh autodetected_org_settings to
                with contextlib.suppress(
                    ClientError, NoCredentialsError, FileNotFoundError
                ):
                    self.autodetected_org_settings = self.boto3_session.client(
                        "organizations"
                    ).describe_organization()["Organization"]
            else:
                self.set_boto3_session()
        else:
            self.set_boto3_session()

        asyncio.run(self.sync_config_aws_org())

    def configuration_wizard_aws_accounts(self):
        while True:
            if self.config.aws and self.config.aws.accounts:
                action = questionary.select(
                    "What would you like to do?",
                    choices=["Go back", "Add AWS Account", "Edit AWS Account"],
                ).unsafe_ask()
                if action == "Go back":
                    return
                elif action == "Add AWS Account":
                    self.configuration_wizard_aws_account_add()
                elif action == "Edit AWS Account":
                    self.configuration_wizard_aws_account_edit()
            else:
                self.configuration_wizard_aws_account_add()

    def configuration_wizard_aws_organizations_edit(self):
        org_ids = [org.org_id for org in self.config.aws.organizations]
        org_id_to_config_elem_map = {
            org.org_id: elem for elem, org in enumerate(self.config.aws.organizations)
        }
        if len(org_ids) > 1:
            action = questionary.select(
                "Which AWS Organization would you like to edit?",
                choices=["Go back", *org_ids],
            ).unsafe_ask()
            if action == "Go back":
                return
            org_to_edit = next(
                (org for org in self.config.aws.organizations if org.org_id == action),
                None,
            )
            if not org_to_edit:
                log.debug("Could not find AWS Organization to edit", org_id=action)
                return
        else:
            org_to_edit = self.config.aws.organizations[0]

        choices = [
            "Go back",
            "Update IdentityCenter Region",
            "Update IAMbic control",
            "Update Region",
        ]
        while True:
            action = questionary.select(
                "What would you like to do?",
                choices=choices,
            ).unsafe_ask()
            if action == "Go back":
                return
            elif action == "Update Region":
                org_to_edit.default_region = set_aws_region(
                    "What region should IAMbic use?", org_to_edit.default_region
                )
            elif action == "Update IdentityCenter Region":
                org_to_edit.identity_center = set_identity_center()
                asyncio.run(self.sync_config_aws_org(False))

            confirm_command_exe("AWS Organization", Operation.UPDATED)
            self.config.aws.organizations[
                org_id_to_config_elem_map[org_to_edit.org_id]
            ] = org_to_edit
            self.config.write()

    def configuration_wizard_aws_organizations_add(self):
        if not self.has_cf_permissions:
            log.info(
                "Unable to edit this attribute without CloudFormation permissions."
            )
            return

        org_console_url = f"https://{self.aws_default_region}.console.aws.amazon.com/organizations/v2/home/accounts"
        click.echo(
            f"\nWhat is the AWS Organization ID?\nIt can be found here {org_console_url}"
        )
        org_id = questionary.text(
            "AWS Organization ID: ",
            default=self.autodetected_org_settings.get("Id", ""),
        ).unsafe_ask()
        account_id = self.hub_account_id
        session, profile_name = self.get_boto3_session_for_account(account_id)
        if not session:
            return

        read_only = set_aws_is_read_only()
        self.config.aws.spoke_role_is_read_only = read_only
        assume_as_arn = self.assume_as_arn
        cf_role_arn = self.cf_role_arn

        click.echo(
            "\nIAMbic requires Hub and Spoke roles to be created which is deployed using CloudFormation.\n"
            "To review the templates used or deploy them manually, the templates used can be found here:\n"
            "https://github.com/noqdev/iambic/tree/main/iambic/plugins/v0_1_0/aws/cloud_formation/templates\n"
            "If you have already manually deployed the templates, answer yes to proceed.\n"
            "IAMbic will validate that your stacks have been deployed successfully and will not attempt to replace them."
        )
        if not questionary.confirm("Proceed?").unsafe_ask():
            log.info("Unable to add the AWS Org without creating the required roles.")
            return

        created_successfully = asyncio.run(
            create_iambic_role_stacks(
                cf_client=session.client(
                    "cloudformation", region_name=self.aws_default_region
                ),
                hub_account_id=account_id,
                assume_as_arn=assume_as_arn,
                role_arn=cf_role_arn,
                org_client=session.client(
                    "organizations", region_name=self.aws_default_region
                ),
                read_only=read_only,
            )
        )
        if not created_successfully:
            log.error("Failed to create the required IAMbic roles. Exiting.")
            sys.exit(0)

        aws_org = AWSOrganization(
            org_id=org_id,
            org_account_id=account_id,
            default_rule=BaseAWSOrgRule(),
            hub_role_arn=get_hub_role_arn(account_id),
            aws_profile=profile_name,
            spoke_role_is_read_only=read_only,
            default_region=self.aws_default_region,
        )
        if self.config.aws.spoke_role_is_read_only:
            aws_org.default_rule.iambic_managed = IambicManaged.IMPORT_ONLY
        else:
            aws_org.default_rule.iambic_managed = IambicManaged.READ_AND_WRITE

        self.config.aws.organizations.append(aws_org)

        log.debug("Attempting to get a session on the AWS org", org_id=org_id)
        try:
            session = asyncio.run(aws_org.get_boto3_session())
        except ClientError as e:
            log.error("Unable to get a session on the AWS org", org_id=org_id, error=e)
            session = None

        if (
            session
            and questionary.confirm(
                "Would you like to setup Identity Center (SSO) support?", default=False
            ).unsafe_ask()
        ):
            aws_org.identity_center = set_identity_center()

        if not questionary.confirm("Keep these settings?").unsafe_ask():
            click.echo(
                "\nThe AWS Org will not be added to the config and wizard will exit."
            )
            if questionary.confirm("Proceed?").unsafe_ask():
                log.info("Exiting...")
                sys.exit(0)

        log.info("Saving config.")
        self.config.write()

        if not questionary.confirm(
            "Add the org accounts to the config and import the org's AWS identities?"
        ).unsafe_ask():
            click.echo(
                "\nThis is required to finish the setup process.\nWizard will exit if this has not been setup."
            )
            if questionary.confirm("Exit?").unsafe_ask():
                log.info("Exiting...")
                sys.exit(0)

        asyncio.run(self.sync_config_aws_org())

    def configuration_wizard_aws_organizations(self):
        # Currently only 1 org per config is supported.
        click.echo(
            "\nIf you would like to use AWS Organizations, the IAMbic hub account you configured must be the same "
            "AWS account as your AWS Organization."
        )
        if questionary.confirm("Is this the case?").unsafe_ask():
            if self.config.aws and self.config.aws.organizations:
                self.configuration_wizard_aws_organizations_edit()
            else:
                self.configuration_wizard_aws_organizations_add()

    def configuration_wizard_aws(self):
        self.setup_aws_configuration()
        click.echo(
            "\nWe recommend configuring IAMbic with AWS Organizations, "
            "but you may also manually configure accounts."
        )
        while True:
            action = questionary.select(
                "What would you like to configure in AWS?",
                choices=["Go back", "AWS Organizations", "AWS Accounts"],
            ).unsafe_ask()
            if action == "Go back":
                return
            elif action == "AWS Organizations":
                self.configuration_wizard_aws_organizations()
            elif action == "AWS Accounts":
                self.configuration_wizard_aws_accounts()

    def upsert_secret(self):
        if not self.config.secrets:
            self.config.secrets = {}

        if self.config.okta:
            self.config.secrets["okta"] = get_secret_dict_with_val(
                self.config.okta,
                exclude={"organizations": {"__all__": {"client"}}},
                exclude_unset=True,
            )

        if self.config.azure_ad:
            self.config.secrets["azure_ad"] = get_secret_dict_with_val(
                self.config.azure_ad, exclude_unset=True
            )

        if self.config.google_workspace:
            self.config.secrets["google_workspace"] = get_secret_dict_with_val(
                self.config.google_workspace,
                exclude={"_service_connection_map"},
                exclude_unset=True,
            )

        if len(self.config.extends) == 0:
            self.create_secret()
        else:
            self.update_secret()

    def create_secret(self):
        response = None

        if len(self.config.extends) == 0:
            if self.aws_account_map:
                response = questionary.select(
                    "Where do you want to store your secrets?",
                    choices=[
                        LOCALLY,
                        AWS_SECRETS,
                    ],
                ).ask()
            else:
                # TODO: ask if they want to create the file
                response = LOCALLY
        elif self.config.extends[0].key == ExtendsConfigKey.AWS_SECRETS_MANAGER:
            response = AWS_SECRETS
        elif self.config.extends[0].key == ExtendsConfigKey.LOCAL_FILE:
            response = LOCALLY

        if response == AWS_SECRETS:
            region = set_aws_region(
                "What region should the secret be created in?",
                self.aws_default_region,
            )

            question_text = "Create the secret"
            role_name = IAMBIC_SPOKE_ROLE_NAME
            role_account_id = self.hub_account_id

            if not questionary.confirm(f"{question_text}?").unsafe_ask():
                self.config.secrets = {}
                return

            if role_name and (aws_account := self.aws_account_map.get(role_account_id)):
                session = asyncio.run(aws_account.get_boto3_session(region_name=region))
            else:
                session = boto3.Session(region_name=region)

            client = session.client(service_name="secretsmanager", region_name=region)
            response = client.create_secret(
                Name=f"iambic-config-secrets-{str(uuid.uuid4())}",
                Description="IAMbic managed secret used to store protected config values",
                SecretString=yaml.dump({"secrets": self.config.secrets}),
            )

            self.config.extends = [
                ExtendsConfig(
                    key=ExtendsConfigKey.AWS_SECRETS_MANAGER,
                    value=response["ARN"],
                    assume_role_arn=get_spoke_role_arn(self.hub_account_id),
                )  # type: ignore
            ]
            self.config.write()
        elif response == LOCALLY:
            # TODO: add check to gitignore the file
            with open(self.secret_config_dir(), "w") as f:
                f.write(yaml.dump({"secrets": self.config.secrets}))  # type: ignore

            self.config.extends = [
                ExtendsConfig(
                    key=ExtendsConfigKey.LOCAL_FILE,
                    value="secrets.yaml",
                )  # type: ignore
            ]

            self.config.write()

    def update_secret(self):
        response = None

        if self.config.extends[0].key == ExtendsConfigKey.AWS_SECRETS_MANAGER:
            response = AWS_SECRETS
        elif self.config.extends[0].key == ExtendsConfigKey.LOCAL_FILE:
            response = LOCALLY

        secret_details = self.config.extends[0]

        if response == AWS_SECRETS:
            secret_arn = secret_details.value
            region = secret_arn.split(":")[3]
            secret_account_id = secret_arn.split(":")[4]

            if aws_account := self.aws_account_map.get(secret_account_id):
                session = asyncio.run(aws_account.get_boto3_session(region_name=region))
            else:
                session = boto3.Session(region_name=region)

            client = session.client(service_name="secretsmanager", region_name=region)
            client.put_secret_value(
                SecretId=secret_arn,
                SecretString=yaml.dump({"secrets": self.config.secrets}),
            )
        elif response == LOCALLY:
            secrets_config_path = self.secret_config_dir(secret_details.value)
            with open(secrets_config_path, "w") as f:
                f.write(yaml.dump({"secrets": self.config.secrets}))  # type: ignore

            self.config.write()

    def configuration_wizard_google_workspace_add(self):
        google_obj = {
            "subjects": [set_google_subject()],
            "type": set_google_project_type(),
            "project_id": set_google_project_id(),
            "private_key_id": set_google_private_key_id(),
            "private_key": set_google_private_key(),
            "client_email": set_google_client_email(),
            "client_id": set_client_id(),
            "auth_uri": set_google_auth_uri(),
            "token_uri": set_google_token_uri(),
            "auth_provider_x509_cert_url": set_google_auth_provider_cert_url(),
            "client_x509_cert_url": set_google_client_cert_url(),
        }

        confirm_command_exe("Google Workspace", Operation.ADDED)

        project = GoogleProject(**google_obj)

        if self.config.secrets and self.config.google_workspace:
            self.config.google_workspace.workspaces.append(project)
        else:
            self.config.google_workspace = GoogleWorkspaceConfig(workspaces=[project])

        asyncio.run(self.run_import_google_workspace_resources())
        self.upsert_secret()

    def configuration_wizard_google_workspace_edit(self):
        project_ids = [
            project.project_id for project in self.config.google_workspace.workspaces
        ]
        project_id_to_config_elem_map = {
            project.project_id: elem
            for elem, project in enumerate(self.config.google_workspace.workspaces)
        }
        if len(project_ids) > 1:
            action = questionary.select(
                "Which Google Workspace would you like to edit?",
                choices=["Go back", *project_ids],
            ).unsafe_ask()
            if action == "Go back":
                return
            project_to_edit = next(
                (
                    project
                    for project in self.config.google_workspace.workspaces
                    if project.project_id == action
                ),
                None,
            )
            if not project_to_edit:
                log.debug("Could not find AWS Organization to edit", org_id=action)
                return
        else:
            project_to_edit = self.config.google_workspace.workspaces[0]

        project_id = project_to_edit.project_id
        choices = [
            "Go back",
            "Update Subject",
            "Update Type",
            "Update Private Key",
            "Update Private Key ID",
            "Update Client Email",
            "Update Client ID",
            "Update Auth URI",
            "Update Token URI",
            "Update Auth Provider Cert URL",
            "Update Client Cert URL",
        ]
        while True:
            action = questionary.select(
                "What would you like to do?",
                choices=choices,
            ).unsafe_ask()
            if action == "Go back":
                return
            elif action == "Update Subject":
                if project_to_edit.subjects:
                    default_domain = project_to_edit.subjects[0].domain
                    default_service = project_to_edit.subjects[0].service_account
                else:
                    default_domain = None
                    default_service = None
                project_to_edit.subjects = [
                    GoogleSubject(**set_google_subject(default_domain, default_service))
                ]
            elif action == "Update Type":
                project_to_edit.type = set_google_project_type(project_to_edit.type)
            elif action == "Update Private Key":
                project_to_edit.private_key = set_google_private_key(
                    project_to_edit.private_key
                )
            elif action == "Update Private Key ID":
                project_to_edit.private_key_id = set_google_private_key_id(
                    project_to_edit.private_key_id
                )
            elif action == "Update Client Email":
                project_to_edit.client_email = set_google_client_email(
                    project_to_edit.client_email
                )
            elif action == "Update Client ID":
                project_to_edit.client_id = set_client_id(project_to_edit.client_id)
            elif action == "Update Auth URI":
                project_to_edit.auth_uri = set_google_auth_uri(project_to_edit.auth_uri)
            elif action == "Update Token URI":
                project_to_edit.token_uri = set_google_token_uri(
                    project_to_edit.token_uri
                )
            elif action == "Update Auth Provider Cert URL":
                project_to_edit.auth_provider_x509_cert_url = (
                    set_google_auth_provider_cert_url(
                        project_to_edit.auth_provider_x509_cert_url
                    )
                )
            elif action == "Update Client Cert URL":
                project_to_edit.client_x509_cert_url = set_google_client_cert_url(
                    project_to_edit.client_x509_cert_url
                )

            confirm_command_exe("Google Workspace", Operation.UPDATED)

            self.config.google_workspace.workspaces[
                project_id_to_config_elem_map[project_id]
            ] = project_to_edit

            asyncio.run(self.run_import_google_workspace_resources())
            self.upsert_secret()
            self.config.write()

    def configuration_wizard_google_workspace(self):
        log.info(
            "For details on how to retrieve the information required to add a Google Workspace "
            "to IAMbic check out our docs: https://docs.iambic.org/getting_started/google/"
        )

        if self.config.google_workspace:
            action = questionary.select(
                "What would you like to do?",
                choices=["Go back", "Add", "Edit"],
            ).unsafe_ask()
            if action == "Go back":
                return
            elif action == "Add":
                self.configuration_wizard_google_workspace_add()
            elif action == "Edit":
                self.configuration_wizard_google_workspace_edit()
        else:
            self.configuration_wizard_google_workspace_add()

    def configuration_wizard_okta_organization_add(self):
        okta_obj = {
            "idp_name": set_idp_name(),
            "org_url": set_okta_org_url(),
            "api_token": set_okta_api_token(),
        }

        confirm_command_exe("Okta Organization", Operation.ADDED)

        organization = OktaOrganization(**okta_obj)

        if self.config.secrets and self.has_okta_configured:
            self.config.okta.organizations.append(organization)
        else:
            self.config.okta = OktaConfig(organizations=[organization])

        asyncio.run(self.run_import_okta_resources())
        self.upsert_secret()

    def configuration_wizard_okta_organization_edit(self):
        org_names = [org.idp_name for org in self.config.okta.organizations]
        org_name_to_config_elem_map = {
            org.idp_name: elem
            for elem, org in enumerate(self.config.okta.organizations)
        }
        if len(org_names) > 1:
            action = questionary.select(
                "Which Okta Organization would you like to edit?",
                choices=["Go back", *org_names],
            ).unsafe_ask()
            if action == "Go back":
                return
            org_to_edit = next(
                (
                    org
                    for org in self.config.okta.organizations
                    if org.idp_name == action
                ),
                None,
            )
            if not org_to_edit:
                log.debug("Could not find Okta Organization to edit", idp_name=action)
                return
        else:
            org_to_edit = self.config.okta.organizations[0]

        org_name = org_to_edit.idp_name
        choices = [
            "Go back",
            "Update name",
            "Update Organization URL",
            "Update API Token",
        ]
        while True:
            action = questionary.select(
                "What would you like to do?",
                choices=choices,
            ).unsafe_ask()
            if action == "Go back":
                return
            elif action == "Update name":
                org_to_edit.idp_name = set_idp_name(org_to_edit.idp_name)
            elif action == "Update Organization URL":
                org_to_edit.org_url = set_okta_org_url(org_to_edit.org_url)
            elif action == "Update API Token":
                org_to_edit.api_token = set_okta_api_token(org_to_edit.api_token)

            confirm_command_exe("Okta Organization", Operation.UPDATED)
            self.config.okta.organizations[
                org_name_to_config_elem_map[org_name]
            ] = org_to_edit

            asyncio.run(self.run_import_okta_resources())
            self.upsert_secret()
            self.config.write()

    def configuration_wizard_okta(self):
        log.info(
            "For details on how to retrieve the information required to add an Okta Organization "
            "to IAMbic check out our docs: https://docs.iambic.org/getting_started/okta/"
        )
        if self.config.okta:
            action = questionary.select(
                "What would you like to do?",
                choices=["Go back", "Add", "Edit"],
            ).unsafe_ask()
            if action == "Go back":
                return
            elif action == "Add":
                self.configuration_wizard_okta_organization_add()
            elif action == "Edit":
                self.configuration_wizard_okta_organization_edit()
        else:
            self.configuration_wizard_okta_organization_add()

    def configuration_wizard_azure_ad_organization_add(self):
        azure_ad_obj = {
            "idp_name": set_idp_name(),
            "tenant_id": set_tenant_id(),
            "client_id": set_client_id(),
            "client_secret": set_client_secret(),
        }

        confirm_command_exe("Azure AD Organization", Operation.ADDED)

        organization = AzureADOrganization(**azure_ad_obj)

        if (
            self.config.secrets
            and self.config.azure_ad
            and self.config.azure_ad.organizations
        ):
            self.config.azure_ad.organizations.append(organization)
        else:
            self.config.azure_ad = AzureADConfig(organizations=[organization])

        asyncio.run(self.run_import_azure_ad_resources())
        self.upsert_secret()

    def configuration_wizard_azure_ad_organization_edit(self):
        org_names = [org.idp_name for org in self.config.azure_ad.organizations]
        org_name_to_config_elem_map = {
            org.idp_name: elem
            for elem, org in enumerate(self.config.azure_ad.organizations)
        }
        if len(org_names) > 1:
            action = questionary.select(
                "Which Azure AD Organization would you like to edit?",
                choices=["Go back", *org_names],
            ).unsafe_ask()
            if action == "Go back":
                return
            org_to_edit = next(
                (
                    org
                    for org in self.config.azure_ad.organizations
                    if org.idp_name == action
                ),
                None,
            )
            if not org_to_edit:
                log.debug(
                    "Could not find Azure AD Organization to edit", idp_name=action
                )
                return
        else:
            org_to_edit = self.config.azure_ad.organizations[0]

        org_name = org_to_edit.idp_name
        choices = [
            "Go back",
            "Update IDP name",
            "Update Tenant ID",
            "Update Client ID",
            "Update Client Secret",
        ]
        while True:
            action = questionary.select(
                "What would you like to do?",
                choices=choices,
            ).unsafe_ask()
            if action == "Go back":
                return
            elif action == "Update IDP name":
                org_to_edit.idp_name = set_idp_name(org_to_edit.idp_name)
            elif action == "Update Tenant ID":
                org_to_edit.tenant_id = set_tenant_id(org_to_edit.tenant_id)
            elif action == "Update Client ID":
                org_to_edit.client_id = set_client_id(org_to_edit.client_id)
            elif action == "Update Client Secret":
                org_to_edit.client_secret = set_client_secret()

            confirm_command_exe("Azure AD Organization", Operation.UPDATED)
            self.config.azure_ad.organizations[
                org_name_to_config_elem_map[org_name]
            ] = org_to_edit

            asyncio.run(self.run_import_azure_ad_resources())

            self.upsert_secret()
            self.config.write()

    def configuration_wizard_azure_ad(self):
        log.info(
            "For details on how to retrieve the information required to add an Azure AD Organization "
            "to IAMbic check out our docs: https://docs.iambic.org/getting_started/azure_ad/"
        )
        if self.config.azure_ad:
            action = questionary.select(
                "What would you like to do?",
                choices=["Go back", "Add", "Edit"],
            ).unsafe_ask()
            if action == "Go back":
                return
            elif action == "Add":
                self.configuration_wizard_azure_ad_organization_add()
            elif action == "Edit":
                self.configuration_wizard_azure_ad_organization_edit()
        else:
            self.configuration_wizard_azure_ad_organization_add()

    def configuration_wizard_github_workflow(self):
        log.info(
            "NOTE: Currently, only GitHub Workflows are supported. "
            "However, you can modify the generated output to work with your Git provider."
        )

        if questionary.confirm("Proceed?").unsafe_ask():
            commit_email = set_required_text_value(
                "What is the E-Mail address to use for commits?"
            )
            repo_name = set_required_text_value(
                "What is the name of the repository, including the organization (example: github_org/repo_name)?"
            )
            if self.config.aws and self.config.aws.organizations:
                aws_org = self.config.aws.organizations[0]
                region = aws_org.region_name
            else:
                region = set_aws_region(
                    "What AWS region should the workflow run in?",
                    default_val=RegionName.us_east_1,
                )

            create_workflow_files(
                repo_dir=self.repo_dir,
                repo_name=repo_name,
                commit_email=commit_email,
                assume_role_arn=self.config.aws.hub_role_arn,
                region=region,
            )

    def configuration_wizard_change_detection_setup(self, aws_org: AWSOrganization):
        click.echo(
            "\nTo setup change detection for iambic it requires creating CloudFormation stacks "
            "and a CloudFormation stack set.\n"
            "To review the templates used or deploy them manually, the IdentityRule templates used can be found here:\n"
            "https://github.com/noqdev/iambic/tree/main/iambic/plugins/v0_1_0/aws/cloud_formation/templates\n"
            "If you have already manually deployed the templates, answer yes to proceed.\n"
            "IAMbic will validate that your stacks have been deployed successfully and will not attempt to replace them."
        )
        if not questionary.confirm("Proceed?").unsafe_ask():
            return

        session, _ = self.get_boto3_session_for_account(aws_org.org_account_id)
        cf_client = session.client(
            "cloudformation", region_name=self.aws_default_region
        )
        org_client = session.client(
            "organizations", region_name=self.aws_default_region
        )

        successfully_created = asyncio.run(
            create_iambic_eventbridge_stacks(
                cf_client,
                org_client,
                aws_org.org_id,
                aws_org.org_account_id,
                self.cf_role_arn,
            )
        )
        if not successfully_created:
            return

        role_name = IAMBIC_SPOKE_ROLE_NAME
        hub_account_id = self.hub_account_id
        sqs_arn = f"arn:aws:sqs:{self.aws_default_region}:{hub_account_id}:IAMbicChangeDetectionQueue"

        if not self.existing_role_template_map:
            log.info("Loading AWS role templates...")
            self.existing_role_template_map = asyncio.run(
                get_existing_template_map(self.repo_dir, AWS_IAM_ROLE_TEMPLATE_TYPE)
            )

        role_template: AwsIamRoleTemplate = self.existing_role_template_map.get(
            role_name
        )

        self.config.aws.sqs_cloudtrail_changes_queues = [sqs_arn]
        asyncio.run(self.save_and_deploy_changes(role_template))

    def run(self):  # noqa: C901
        """Run the configuration wizard.

        This will prompt the user for the required information to configure IAMbic.

        Notes:
        - AWS is not required to use IAMbic, but it is required to store to aws secrets.
        """

        while True:
            choices = []
            secret_question_text, _ = self.secrets_message

            if "aws" in self.config.__fields__:
                choices.append("AWS")
            if "azure_ad" in self.config.__fields__:
                choices.append("Azure AD")
            if "google_workspace" in self.config.__fields__:
                choices.append("Google Workspace")
            if "okta" in self.config.__fields__:
                choices.append("Okta")

            if self.has_aws_account_or_organizations:
                choices.append("Generate Github Action Workflows")
                if (
                    self.config.aws.organizations
                    and not self.config.aws.sqs_cloudtrail_changes_queues
                ):
                    choices.append("Setup AWS change detection")

            choices.append("Done")

            try:
                action = questionary.select(
                    "What would you like to configure?",
                    choices=choices,
                ).unsafe_ask()
            except KeyboardInterrupt:
                log.info("Exiting...")
                return

            # Let's try really hard not to use a switch statement since it depends on Python 3.10
            try:
                if action == "Done":
                    self.config.write()
                    return
                elif action == "AWS":
                    self.configuration_wizard_aws()
                elif action == "Google Workspace":
                    if questionary.confirm(
                        f"{secret_question_text} Proceed?"
                    ).unsafe_ask():
                        self.configuration_wizard_google_workspace()
                elif action == "Okta":
                    if questionary.confirm(
                        f"{secret_question_text} Proceed?"
                    ).unsafe_ask():
                        self.configuration_wizard_okta()
                elif action == "Azure AD":
                    if questionary.confirm(
                        f"{secret_question_text} Proceed?"
                    ).unsafe_ask():
                        self.configuration_wizard_azure_ad()
                elif action == "Generate Github Action Workflows":
                    self.configuration_wizard_github_workflow()
                elif action == "Setup AWS change detection":
                    if self.has_cf_permissions:
                        log.info(
                            f"IAMbic change detection relies on CloudTrail being enabled all IAMbic aware accounts. "
                            f"You can check that you have CloudTrail setup by going to "
                            f"https://{self.aws_default_region}.console.aws.amazon.com/cloudtrail/home\n"
                            f"If you do not have CloudTrail setup, you can set it up by going to "
                            f"https://{self.aws_default_region}.console.aws.amazon.com/cloudtrail/home?region={self.aws_default_region}#/create"
                        )
                        self.configuration_wizard_change_detection_setup(
                            self.config.aws.organizations[0]
                        )
                    else:
                        log.info(
                            "Unable to edit this attribute without CloudFormation permissions."
                        )
            except KeyboardInterrupt:
                ...


monkeypatch_questionary()



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.example.iambic_plugin import ExampleConfig


async def load(config: ExampleConfig) -> ExampleConfig:
    return config


async def import_example_resources(
    config: ExampleConfig, base_output_dir: str, messages: list = None
):
    tasks = []
    await asyncio.gather(*tasks)



from __future__ import annotations

import unittest

import pytest
from pydantic import ValidationError

from iambic.plugins.v0_1_0.aws.iam.models import MaxSessionDuration, Path
from iambic.plugins.v0_1_0.aws.iam.role.models import AwsIamRoleTemplate, RoleProperties
from iambic.plugins.v0_1_0.aws.iambic_plugin import AWSConfig
from iambic.plugins.v0_1_0.aws.models import AWSAccount, Description, Partition


def test_unique_account_id():
    with pytest.raises(ValidationError, match=r"1 validation error for AWSConfig.*"):
        AWSConfig(
            accounts=[
                AWSAccount(account_id="123456789012", account_name="account1"),
                AWSAccount(account_id="123456789012", account_name="account2"),
            ]
        )


def test_unique_account_name():
    with pytest.raises(ValidationError, match=r"1 validation error for AWSConfig.*"):
        AWSConfig(
            accounts=[
                AWSAccount(account_id="123456789012", account_name="account1"),
                AWSAccount(account_id="234567890123", account_name="account1"),
            ]
        )


def test_unique_account_id_and_name():
    with pytest.raises(ValidationError, match=r"1 validation error for AWSAccount.*"):
        AWSConfig(
            accounts=[
                AWSAccount(account_id="123456789012"),
                AWSAccount(account_id="123456789012"),
            ]
        )


def test_valid_config():
    config = AWSConfig(
        accounts=[
            AWSAccount(account_id="123456789012", account_name="account1"),
            AWSAccount(account_id="234567890123", account_name="account2"),
        ]
    )
    assert config.validate(config)


class TestBaseModel(unittest.TestCase):
    def setUp(self):
        properties = RoleProperties(
            role_name="test_role",
            description=[
                Description(
                    included_accounts=["test_account"],
                    description="test_account role description",
                ),
                Description(
                    included_accounts=["not_test_account"],
                    description="not_test_account role description",
                ),
            ],
            max_session_duration=[
                MaxSessionDuration(
                    included_accounts=["test_account"], max_session_duration=3600
                ),
                MaxSessionDuration(
                    included_accounts=["not_test_account"], max_session_duration=600
                ),
            ],
            path=[Path(file_path="/")],
        )
        self.role_template = AwsIamRoleTemplate(
            identifier="test_role",
            file_path="/tmp/test_role.yaml",
            properties=properties,
        )

    def test_get_attribute_val_for_account(self):
        aws_account = AWSAccount(
            account_id="123456789012",
            account_name="test_account",
            partition=Partition.AWS,
        )

        result = self.role_template.get_attribute_val_for_account(
            aws_account, "properties.role_name"
        )
        self.assertEqual(result, "test_role")

        result = self.role_template.get_attribute_val_for_account(
            aws_account, "properties.description"
        )
        self.assertEqual(result, [{"Description": "test_account role description"}])

        result = self.role_template.get_attribute_val_for_account(
            aws_account, "properties.max_session_duration"
        )
        self.assertEqual(result, [{"MaxSessionDuration": 3600}])



from __future__ import annotations

from iambic.core.iambic_enum import Command


class ExecutionContext:
    eval_only: bool = False
    use_remote: bool = False
    command: Command = None

    @property
    def execute(self):
        return not self.eval_only


ctx = ExecutionContext()



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from iambic.core.context import ctx
from iambic.core.iambic_enum import IambicManaged
from iambic.core.logger import log
from iambic.core.models import ExecutionMessage
from iambic.plugins.v0_1_0.okta.app.template_generation import (
    collect_org_apps,
    generate_app_templates,
)
from iambic.plugins.v0_1_0.okta.group.template_generation import (
    collect_org_groups,
    generate_group_templates,
)
from iambic.plugins.v0_1_0.okta.user.template_generation import (
    collect_org_users,
    generate_user_templates,
)

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.okta.iambic_plugin import OktaConfig


async def load(config: OktaConfig) -> OktaConfig:
    return config


async def import_okta_resources(
    exe_message: ExecutionMessage,
    config: OktaConfig,
    base_output_dir: str,
    messages: list = None,
    remote_worker=None,
):
    base_runner = bool(not exe_message.provider_id)
    collector_tasks = []

    for organization in config.organizations:
        if organization.iambic_managed == IambicManaged.DISABLED:
            continue
        elif (
            exe_message.provider_id and exe_message.provider_id != organization.idp_name
        ):
            continue

        task_message = exe_message.copy()
        task_message.provider_id = organization.idp_name

        collector_tasks.extend(
            [
                collect_org_apps(task_message, config),
                collect_org_groups(task_message, config),
                collect_org_users(task_message, config),
            ]
        )

    if collector_tasks:
        if base_runner and ctx.use_remote and remote_worker and not messages:
            # TODO: Update to use the remote_worker
            await asyncio.gather(*collector_tasks)
            # TODO: Add a process to gather status messages from the remote worker
        else:
            if remote_worker:
                log.warning(
                    "The remote worker definition must be defined in the config to run remote execution."
                )
            await asyncio.gather(*collector_tasks)

    if base_runner:
        generator_tasks = [
            generate_app_templates(exe_message, base_output_dir),
            generate_group_templates(exe_message, base_output_dir),
            generate_user_templates(exe_message, base_output_dir),
        ]
        await asyncio.gather(*generator_tasks)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from pydantic import Field

from iambic.core.models import (
    BaseModel,
    BaseTemplate,
    ExpiryModel,
    ProposedChange,
    ProposedChangeType,
    TemplateChangeDetails,
)

EXAMPLE_LOCAL_FILE_TEMPLATE_TYPE = "NOQ::Example::LocalFile"
EXAMPLE_LOCAL_FILE_MULTI_ACCOUNT_TEMPLATE_TYPE = "NOQ::Example::LocalFileMultiAccount"

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.example.iambic_plugin import ExampleConfig


class ExampleLocalFileTemplateProperties(BaseModel):
    name: str = Field(..., description="name of local file")

    @property
    def resource_type(self) -> str:
        return "example:local_file:properties"

    @property
    def resource_id(self) -> str:
        return self.name


class ExampleLocalFileTemplate(BaseTemplate, ExpiryModel):
    template_type = EXAMPLE_LOCAL_FILE_TEMPLATE_TYPE
    properties: ExampleLocalFileTemplateProperties = Field(
        ..., description="Properties for Example Local File Template"
    )
    name: str = Field(..., description="name of local file")

    @property
    def resource_type(self) -> str:
        return "example:local_file"

    @property
    def resource_id(self) -> str:
        return self.name

    async def apply(self, config: ExampleConfig) -> TemplateChangeDetails:
        template_changes = TemplateChangeDetails(
            resource_id=self.resource_id,
            resource_type=self.template_type,
            template_path=self.file_path,
        )
        if self.delete:
            template_changes.proposed_changes = [
                ProposedChange(change_type=ProposedChangeType.DELETE)
            ]
        else:
            template_changes.proposed_changes = []
        return template_changes


class ExampleLocalFileMultiAccountTemplateProperties(BaseTemplate, ExpiryModel):
    name: str = Field(..., description="name of local file")

    @property
    def resource_type(self) -> str:
        return "example:local_file_multi_account:properties"

    @property
    def resource_id(self) -> str:
        return self.name


class ExampleLocalFileMultiAccountTemplate(BaseTemplate, ExpiryModel):
    template_type = EXAMPLE_LOCAL_FILE_MULTI_ACCOUNT_TEMPLATE_TYPE
    properties: ExampleLocalFileTemplateProperties = Field(
        ..., description="Properties for Example Local File Template"
    )
    name: str = Field(..., description="name of local file")
    included_accounts: Optional[list] = Field(
        ..., description="list of accounts to include"
    )
    excluded_accounts: Optional[list] = Field(
        ..., description="list of accounts to exclude"
    )

    @property
    def resource_type(self) -> str:
        return "example:local_file_multi_account"

    @property
    def resource_id(self) -> str:
        return self.name

    async def apply(self, config: ExampleConfig) -> TemplateChangeDetails:
        template_changes = TemplateChangeDetails(
            resource_id=self.resource_id,
            resource_type=self.template_type,
            template_path=self.file_path,
        )
        if self.delete:
            template_changes.proposed_changes = [
                ProposedChange(change_type=ProposedChangeType.DELETE)
            ]
        else:
            template_changes.proposed_changes = []
        return template_changes



from __future__ import annotations

import unittest
from unittest import IsolatedAsyncioTestCase
from unittest.mock import AsyncMock, MagicMock

from iambic.core.models import ProposedChange, ProposedChangeType
from iambic.plugins.v0_1_0.google_workspace.group.models import (
    GoogleWorkspaceGroupTemplate,
    GroupMember,
    GroupProperties,
)
from iambic.plugins.v0_1_0.google_workspace.group.utils import (
    create_group,
    list_groups,
    maybe_delete_group,
    update_group_description,
    update_group_domain,
    update_group_email,
    update_group_members,
    update_group_name,
)
from iambic.plugins.v0_1_0.google_workspace.iambic_plugin import GoogleProject
from iambic.plugins.v0_1_0.google_workspace.models import (
    GroupMemberRole,
    GroupMemberStatus,
    GroupMemberType,
)


class TestListGroups(IsolatedAsyncioTestCase):
    async def test_list_groups(self):
        # Set up mock data
        domain = "example.com"
        group_data = [
            {"id": "group1@example.com", "email": "group1@example.com"},
            {"id": "group2@example.com", "email": "group2@example.com"},
        ]

        # Set up mock service connection
        mock_service = MagicMock()
        mock_service.groups.return_value.list.return_value.execute.return_value = {
            "groups": group_data,
        }
        google_project = MagicMock()
        google_project.get_service_connection = AsyncMock(return_value=mock_service)

        # Set up mock get_group_template function

        with unittest.mock.patch(
            "iambic.plugins.v0_1_0.google_workspace.group.models.get_group_template",
            new_callable=AsyncMock,
        ) as mock_get_group_template:
            mock_get_group_template.side_effect = lambda service, group, domain: group

            # Call list_groups function
            result = await list_groups(domain, google_project)

            # Check if the result is as expected
            self.assertEqual(len(result), len(group_data))
            for idx, group in enumerate(result):
                self.assertEqual(group["id"], group_data[idx]["id"])
                self.assertEqual(group["email"], group_data[idx]["email"])

            # Check if the get_group_template function was called for each group
            self.assertEqual(mock_get_group_template.call_count, len(group_data))


class TestCreateGroup(IsolatedAsyncioTestCase):
    async def test_create_group(self):
        # Set up mock data
        domain = "example.com"
        group_id = "group1@example.com"
        group_email = "group1@example.com"
        group_name = "Group 1"
        group_description = "A test group"

        # Set up mock service connection
        mock_service = MagicMock()
        mock_service.groups.return_value.insert.return_value.execute.return_value = {
            "id": group_id,
            "email": group_email,
            "name": group_name,
            "description": group_description,
        }
        google_project = MagicMock()
        google_project.get_service_connection = AsyncMock(return_value=mock_service)

        # Call create_group function
        result = await create_group(
            id=group_id,
            domain=domain,
            email=group_email,
            name=group_name,
            description=group_description,
            google_project=google_project,
        )

        # Check if the result is as expected
        self.assertEqual(result["id"], group_id)
        self.assertEqual(result["email"], group_email)
        self.assertEqual(result["name"], group_name)
        self.assertEqual(result["description"], group_description)

        # Check if the insert function was called with the correct parameters
        mock_service.groups.return_value.insert.assert_called_with(
            body={
                "id": group_id,
                "email": group_email,
                "name": group_name,
                "description": group_description,
            }
        )


class TestUpdateGroupDomain(IsolatedAsyncioTestCase):
    async def test_update_group_domain_same_domain(self):
        current_domain = "example.com"
        proposed_domain = "example.com"
        log_params = {
            "resource_type": "google:group",
            "resource_id": "group1@example.com",
            "account": "example.com",
        }

        result = await update_group_domain(current_domain, proposed_domain, log_params)

        self.assertEqual(result, [])

    async def test_update_group_domain_different_domain(self):
        current_domain = "example.com"
        proposed_domain = "newexample.com"
        log_params = {
            "resource_type": "google:group",
            "resource_id": "group1@example.com",
            "account": "example.com",
        }
        with self.assertRaises(NotImplementedError):
            await update_group_domain(current_domain, proposed_domain, log_params)


class TestUpdateGroupDescription(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.google_project = MagicMock(spec=GoogleProject)
        self.google_project.get_service_connection = AsyncMock(return_value=MagicMock())
        self.log_params = {
            "resource_type": "google:group",
            "resource_id": "group1@example.com",
            "account": "example.com",
        }

    async def test_update_group_description_same_description(self):
        group_email = "group1@example.com"
        current_description = "Current Group Description"
        proposed_description = "Current Group Description"
        domain = "example.com"

        result = await update_group_description(
            group_email,
            current_description,
            proposed_description,
            domain,
            self.google_project,
            self.log_params,
        )

        self.assertEqual(result, [])

    async def test_update_group_description_different_description(self):
        group_email = "group1@example.com"
        current_description = "Current Group Description"
        proposed_description = "New Group Description"
        domain = "example.com"

        result = await update_group_description(
            group_email,
            current_description,
            proposed_description,
            domain,
            self.google_project,
            self.log_params,
        )

        self.assertEqual(len(result), 1)
        self.assertIsInstance(result[0], ProposedChange)
        self.assertEqual(result[0].change_type, ProposedChangeType.UPDATE)
        self.assertEqual(result[0].attribute, "description")
        self.assertEqual(
            result[0].change_summary,
            {
                "current_description": current_description,
                "proposed_description": proposed_description,
            },
        )


class TestUpdateGroupName(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.google_project = MagicMock(spec=GoogleProject)
        self.google_project.get_service_connection = AsyncMock(return_value=MagicMock())
        self.log_params = {
            "resource_type": "google:group",
            "resource_id": "group1@example.com",
            "account": "example.com",
        }

    async def test_update_group_name_same_name(self):
        group_email = "group1@example.com"
        current_name = "Current Group Name"
        proposed_name = "Current Group Name"
        domain = "example.com"

        result = await update_group_name(
            group_email,
            current_name,
            proposed_name,
            domain,
            self.google_project,
            self.log_params,
        )

        self.assertEqual(result, [])

    async def test_update_group_name_different_name(self):
        group_email = "group1@example.com"
        current_name = "Current Group Name"
        proposed_name = "New Group Name"
        domain = "example.com"

        result = await update_group_name(
            group_email,
            current_name,
            proposed_name,
            domain,
            self.google_project,
            self.log_params,
        )

        self.assertEqual(len(result), 1)
        self.assertIsInstance(result[0], ProposedChange)
        self.assertEqual(result[0].change_type, ProposedChangeType.UPDATE)
        self.assertEqual(result[0].attribute, "group_name")
        self.assertEqual(result[0].new_value, proposed_name)


class TestUpdateGroupEmail(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.google_project = MagicMock(spec=GoogleProject)
        self.google_project.get_service_connection = AsyncMock(return_value=MagicMock())
        self.log_params = {
            "resource_type": "google:group",
            "resource_id": "group1@example.com",
            "account": "example.com",
        }

    async def test_update_group_email_same_email(self):
        current_email = "group1@example.com"
        proposed_email = "group1@example.com"
        domain = "example.com"

        result = await update_group_email(
            current_email,
            proposed_email,
            domain,
            self.google_project,
            self.log_params,
        )

        self.assertEqual(result, [])

    async def test_update_group_email_different_email(self):
        current_email = "group1@example.com"
        proposed_email = "group2@example.com"
        domain = "example.com"

        result = await update_group_email(
            current_email,
            proposed_email,
            domain,
            self.google_project,
            self.log_params,
        )

        self.assertEqual(len(result), 1)
        self.assertIsInstance(result[0], ProposedChange)
        self.assertEqual(result[0].change_type, ProposedChangeType.UPDATE)
        self.assertEqual(result[0].attribute, "group_email")
        self.assertEqual(result[0].new_value, proposed_email)


class TestMaybeDeleteGroup(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.google_project = MagicMock(spec=GoogleProject)
        self.google_project.get_service_connection = AsyncMock(return_value=MagicMock())
        self.log_params = {
            "resource_type": "google:group",
            "resource_id": "group1@example.com",
            "account": "example.com",
        }

    async def test_maybe_delete_group_not_deleted(self):
        group = GoogleWorkspaceGroupTemplate(
            properties=GroupProperties(
                email="group1@example.com",
                name="Group 1",
                domain="example.com",
                description="test_group",
                members=[],
            ),
            deleted=False,
            file_path="test_google_group.yaml",
        )

        result = await maybe_delete_group(
            group,
            self.google_project,
            self.log_params,
        )

        self.assertEqual(result, [])

    async def test_maybe_delete_group_deleted(self):
        group = GoogleWorkspaceGroupTemplate(
            file_path="test_google_group.yaml",
            properties=GroupProperties(
                email="group1@example.com",
                name="Group 1",
                domain="example.com",
                description="test_group",
                members=[],
            ),
            deleted=True,
        )
        result = await maybe_delete_group(
            group,
            self.google_project,
            self.log_params,
        )

        self.assertEqual(len(result), 1)
        self.assertIsInstance(result[0], ProposedChange)
        self.assertEqual(result[0].change_type, ProposedChangeType.DELETE)
        self.assertEqual(result[0].attribute, "group")
        self.assertEqual(result[0].change_summary, {"group": group.properties.name})


class TestUpdateGroupMembers(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.google_project = MagicMock(spec=GoogleProject)
        self.google_project.get_service_connection = AsyncMock(return_value=MagicMock())
        self.log_params = {
            "resource_type": "google:group",
            "resource_id": "group1@example.com",
            "account": "example.com",
        }

    async def test_update_group_members_no_changes(self):
        group_email = "group1@example.com"
        current_members = [
            GroupMember(
                email="user1@example.com",
                role=GroupMemberRole.MEMBER,
                type=GroupMemberType.USER,
                status=GroupMemberStatus.ACTIVE,
            ),
        ]
        proposed_members = current_members

        result = await update_group_members(
            group_email,
            current_members,
            proposed_members,
            "example.com",
            self.google_project,
            self.log_params,
        )

        self.assertEqual(result, [])

    async def test_update_group_members_add_and_remove(self):
        group_email = "group1@example.com"
        current_members = [
            GroupMember(
                email="user1@example.com",
                role=GroupMemberRole.MEMBER,
                type=GroupMemberType.USER,
                status=GroupMemberStatus.ACTIVE,
            ),
        ]
        proposed_members = [
            GroupMember(
                email="user2@example.com",
                role=GroupMemberRole.MEMBER,
                type=GroupMemberType.USER,
                status=GroupMemberStatus.ACTIVE,
            ),
        ]

        result = await update_group_members(
            group_email,
            current_members,
            proposed_members,
            "example.com",
            self.google_project,
            self.log_params,
        )

        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[0], ProposedChange)
        self.assertEqual(result[0].change_type, ProposedChangeType.DETACH)
        self.assertEqual(result[0].attribute, "users")
        self.assertEqual(
            result[0].change_summary, {"UsersToRemove": ["user1@example.com"]}
        )

        self.assertIsInstance(result[1], ProposedChange)
        self.assertEqual(result[1].change_type, ProposedChangeType.ATTACH)
        self.assertEqual(result[1].attribute, "users")
        self.assertEqual(
            result[1].change_summary, {"UsersToAdd": ["user2@example.com"]}
        )



from __future__ import annotations

import asyncio
import typing
from enum import Enum
from itertools import chain
from typing import Optional

from aiohttp import ClientResponseError
from pydantic import Field

from iambic.core.context import ctx
from iambic.core.logger import log
from iambic.core.models import (
    AccountChangeDetails,
    BaseModel,
    ExpiryModel,
    ProposedChange,
    ProposedChangeType,
)
from iambic.core.utils import normalize_dict_keys
from iambic.plugins.v0_1_0.azure_ad.models import AzureADOrganization, AzureADTemplate

if typing.TYPE_CHECKING:
    MappingIntStrAny = typing.Mapping[int | str, any]
    AbstractSetIntStr = typing.AbstractSet[int | str]

AZURE_AD_USER_TEMPLATE_TYPE = "NOQ::AzureAD::User"


class UserStatus(Enum):
    active = "active"
    provisioned = "provisioned"
    deprovisioned = "deprovisioned"


class UserTemplateProperties(BaseModel, ExpiryModel):
    user_id: Optional[str] = Field("", description="Unique identifier for the user")
    username: str
    display_name: str
    mail_nickname: Optional[str]
    given_name: Optional[str]
    employee_id: Optional[str]
    domain: Optional[str]
    fullname: Optional[str]
    status: Optional[UserStatus]

    business_phones: Optional[list[str]] = Field(
        [], description="List of business phone numbers", exclude=True
    )
    id: Optional[str] = Field(
        None, description="Unique identifier for the user", exclude=True
    )
    job_title: Optional[str] = Field(
        None, description="Job title of the user", exclude=True
    )
    mail: Optional[str] = Field(
        None, description="Email address of the user", exclude=True
    )
    mobile_phone: Optional[str] = Field(
        None, description="Mobile phone number of the user", exclude=True
    )
    office_location: Optional[str] = Field(
        None, description="Office location of the user", exclude=True
    )
    preferred_language: Optional[str] = Field(
        None, description="Preferred language of the user", exclude=True
    )
    surname: Optional[str] = Field(
        None, description="Surname of the user", exclude=True
    )
    user_principal_name: Optional[str] = Field(
        None, description="User principal name of the user", exclude=True
    )

    @property
    def resource_type(self) -> str:
        return "azure_ad:user"

    @property
    def resource_id(self) -> str:
        return self.username

    @classmethod
    def from_azure_response(cls, azure_response: dict) -> UserTemplateProperties:
        unwanted_keys = [
            "password_profile",
            "account_enabled",
        ]
        azure_response = normalize_dict_keys(azure_response)
        # Filter unwanted keys
        azure_response = {
            x: y for x, y in azure_response.items() if x not in unwanted_keys
        }
        azure_response = {
            x: y for x, y in azure_response.items() if not x.startswith("@odata")
        }

        return cls(
            user_id=azure_response.get("id"),
            username=azure_response.get("user_principal_name"),
            **azure_response,
        )


class AzureActiveDirectoryUserTemplate(ExpiryModel, AzureADTemplate):
    template_type = AZURE_AD_USER_TEMPLATE_TYPE
    properties: UserTemplateProperties = Field(
        ..., description="Properties for the Azure AD User"
    )

    @property
    def resource_type(self) -> str:
        return "azure_ad:user"

    async def _apply_to_account(
        self, azure_ad_organization: AzureADOrganization
    ) -> AccountChangeDetails:
        from iambic.plugins.v0_1_0.azure_ad.user.utils import (
            create_user,
            delete_user,
            get_user,
            update_user_attributes,
        )

        change_details = AccountChangeDetails(
            account=self.idp_name,
            resource_id=self.resource_id,
            new_value=self.properties.dict(
                exclude={"metadata_commented_dict"}, exclude_none=True
            ),
            proposed_changes=[],
        )
        cloud_user = None

        log_params = dict(
            resource_type=self.resource_type,
            resource_id=self.resource_id,
            organization=str(self.idp_name),
        )

        if self.properties.user_id and not self.deleted:
            try:
                cloud_user: Optional[UserTemplateProperties] = await get_user(
                    azure_ad_organization,
                    self.properties.user_id,
                    self.properties.username,
                )
                change_details.current_value = cloud_user
            except ClientResponseError as err:
                if err.status == 404:
                    err = (
                        f"User not found in Azure AD where id={self.properties.user_id}"
                    )
                    log.exception(
                        "Invalid user_id provided. User not found in Azure AD.",
                        **log_params,
                    )
                change_details.extend_changes(
                    [
                        ProposedChange(
                            change_type=ProposedChangeType.UPDATE,
                            resource_id=self.resource_id,
                            resource_type=self.resource_type,
                            exceptions_seen=[str(err)],
                        )
                    ]
                )
                return change_details

        user_exists = bool(cloud_user)
        tasks = []

        if not user_exists and not self.deleted:
            log_str = "New resource found in code."
            change_details.extend_changes(
                [
                    ProposedChange(
                        change_type=ProposedChangeType.CREATE,
                        resource_id=self.resource_id,
                        resource_type=self.resource_type,
                    )
                ]
            )
            if not ctx.execute:
                log.info(log_str, **log_params)
                # Exit now because apply functions won't work if resource doesn't exist
                return change_details

            log_str = f"{log_str} Creating resource..."
            log.info(log_str, **log_params)

            try:
                cloud_user: UserTemplateProperties = await create_user(
                    azure_ad_organization=azure_ad_organization,
                    username=self.properties.username,
                    display_name=self.properties.display_name,
                    mail_nickname=self.properties.mail_nickname
                    or self.properties.username.split("@")[0],
                )
                self.properties.user_id = cloud_user.user_id
                self.write()
            except ClientResponseError as err:
                log.exception(
                    "Failed to create user in Azure AD",
                    **log_params,
                )
                proposed_change = change_details.proposed_changes.pop(-1)
                proposed_change.exceptions_seen.append(str(err))
                change_details.exceptions_seen.append(proposed_change)
                return change_details

        if self.deleted:
            change_details.extend_changes(
                await delete_user(
                    azure_ad_organization,
                    self.properties,
                    log_params,
                )
            )
        else:
            tasks.append(
                update_user_attributes(
                    azure_ad_organization,
                    self.properties,
                    cloud_user,
                    log_params,
                )
            )

            changes_made = await asyncio.gather(*tasks)
            if any(changes_made):
                change_details.extend_changes(list(chain.from_iterable(changes_made)))

        if ctx.execute and not change_details.exceptions_seen:
            log.debug(
                "Successfully finished execution for resource",
                changes_made=bool(change_details.proposed_changes),
                **log_params,
            )
        elif change_details.exceptions_seen:
            cmd_verb = "apply" if ctx.execute else "scan for"
            log.error(
                f"Failed to successfully {cmd_verb} resource changes",
                **log_params,
            )
        else:
            log.debug(
                "Successfully finished scanning for drift for resource",
                requires_changes=bool(change_details.proposed_changes),
                **log_params,
            )

        return change_details



from __future__ import annotations

import asyncio
import json
from itertools import chain
from typing import Callable, Optional, Union

from pydantic import Field, validator

from iambic.core.context import ctx
from iambic.core.iambic_enum import Command
from iambic.core.logger import log
from iambic.core.models import (
    AccountChangeDetails,
    BaseModel,
    ExpiryModel,
    ProposedChange,
    ProposedChangeType,
)
from iambic.core.utils import plugin_apply_wrapper
from iambic.plugins.v0_1_0.aws.iam.models import (
    MaxSessionDuration,
    Path,
    PermissionBoundary,
)
from iambic.plugins.v0_1_0.aws.iam.policy.models import (
    AssumeRolePolicyDocument,
    ManagedPolicyRef,
    PolicyDocument,
)
from iambic.plugins.v0_1_0.aws.iam.role.utils import (
    apply_role_inline_policies,
    apply_role_managed_policies,
    apply_role_permission_boundary,
    apply_role_tags,
    delete_iam_role,
    get_role,
    update_assume_role_policy,
)
from iambic.plugins.v0_1_0.aws.models import (
    AccessModel,
    AWSAccount,
    AWSTemplate,
    Description,
    Tag,
)
from iambic.plugins.v0_1_0.aws.utils import boto_crud_call

AWS_IAM_ROLE_TEMPLATE_TYPE = "NOQ::AWS::IAM::Role"


class RoleAccess(ExpiryModel, AccessModel):
    users: list[str] = Field(
        [],
        description="List of users who can assume into the role",
    )
    groups: list[str] = Field(
        [],
        description="List of groups. Users in one or more of the groups can assume into the role",
    )

    @property
    def resource_type(self):
        return "aws:iam:role:access_rule"

    @property
    def resource_id(self):
        return


class RoleProperties(BaseModel):
    role_name: str = Field(
        description="Name of the role",
    )
    description: Optional[Union[str, list[Description]]] = Field(
        "",
        description="Description of the role",
    )
    max_session_duration: Optional[Union[int, list[MaxSessionDuration]]] = 3600
    path: Optional[Union[str, list[Path]]] = "/"
    permissions_boundary: Optional[
        Union[None, PermissionBoundary, list[PermissionBoundary]]
    ] = None
    assume_role_policy_document: Optional[
        Union[None, list[AssumeRolePolicyDocument], AssumeRolePolicyDocument]
    ] = Field(
        [],
        description="Who can assume the Role",
    )
    tags: Optional[list[Tag]] = Field(
        [],
        description="List of tags attached to the role",
    )
    managed_policies: Optional[list[ManagedPolicyRef]] = Field(
        [],
        description="Managed policy arns attached to the role",
    )
    inline_policies: Optional[list[PolicyDocument]] = Field(
        [],
        description="List of the role's inline policies",
    )

    @property
    def resource_type(self):
        return "aws:iam:role"

    @property
    def resource_id(self):
        return self.role_name

    @classmethod
    def sort_func(cls, attribute_name: str) -> Callable:
        def _sort_func(obj):
            return f"{getattr(obj, attribute_name)}!{obj.access_model_sort_weight()}"

        return _sort_func

    @validator("managed_policies")
    def sort_managed_policy_refs(cls, v: list[ManagedPolicyRef]):
        sorted_v = sorted(v, key=cls.sort_func("policy_arn"))
        return sorted_v

    @validator("inline_policies")
    def sort_inline_policies(cls, v: list[PolicyDocument]):
        sorted_v = sorted(v, key=cls.sort_func("policy_name"))
        return sorted_v

    @validator("tags")
    def sort_tags(cls, v: list[Tag]):
        sorted_v = sorted(v, key=cls.sort_func("key"))
        return sorted_v

    @validator("assume_role_policy_document")
    def sort_assume_role_policy_documents(cls, v: list[AssumeRolePolicyDocument]):
        if not isinstance(v, list):
            return v
        sorted_v = sorted(v, key=lambda d: d.access_model_sort_weight())
        return sorted_v

    @validator("max_session_duration")
    def sort_max_session_duration(cls, v: list[MaxSessionDuration]):
        if not isinstance(v, list):
            return v
        sorted_v = sorted(v, key=lambda d: d.access_model_sort_weight())
        return sorted_v

    @validator("permissions_boundary")
    def sort_permissions_boundary(cls, v: list[PermissionBoundary]):
        if not isinstance(v, list):
            return v
        sorted_v = sorted(v, key=lambda d: d.access_model_sort_weight())
        return sorted_v

    @validator("path")
    def sort_path(cls, v: list[Path]):
        if not isinstance(v, list):
            return v
        sorted_v = sorted(v, key=lambda d: d.access_model_sort_weight())
        return sorted_v

    @validator("description")
    def sort_description(cls, v: list[Description]):
        if not isinstance(v, list):
            return v
        sorted_v = sorted(v, key=lambda d: d.access_model_sort_weight())
        return sorted_v


class AwsIamRoleTemplate(AWSTemplate, AccessModel):
    template_type = AWS_IAM_ROLE_TEMPLATE_TYPE
    owner: Optional[str] = Field(None, description="Owner of the role")
    properties: RoleProperties = Field(
        description="Properties of the role",
    )
    access_rules: Optional[list[RoleAccess]] = Field(
        [],
        description="Used to define users and groups who can access the role via Noq credential brokering",
    )

    @validator("access_rules")
    def sort_access_rules(cls, v: list[RoleAccess]):
        sorted_v = sorted(v, key=lambda d: d.access_model_sort_weight())
        return sorted_v

    def _apply_resource_dict(self, aws_account: AWSAccount = None) -> dict:
        response = super(AwsIamRoleTemplate, self)._apply_resource_dict(aws_account)
        response.pop("RoleAccess", None)
        if "Tags" not in response:
            response["Tags"] = []

        # Ensure only 1 of the following objects
        # TODO: Have this handled in a cleaner way. Maybe via an attribute on a pydantic field
        if assume_role_policy := response.pop("AssumeRolePolicyDocument", []):
            if isinstance(assume_role_policy, list):
                assume_role_policy = assume_role_policy[0]
            response["AssumeRolePolicyDocument"] = assume_role_policy

        if permissions_boundary := response.pop("PermissionsBoundary", []):
            if isinstance(permissions_boundary, list):
                permissions_boundary = permissions_boundary[0]
            response["PermissionsBoundary"] = permissions_boundary

        if isinstance(response.get("Description"), list):
            response["Description"] = response["Description"][0]["Description"]

        if isinstance(response.get("MaxSessionDuration"), list):
            response["MaxSessionDuration"] = response["MaxSessionDuration"][0][
                "MaxSessionDuration"
            ]

        return response

    def _is_iambic_import_only(self):
        return "aws-service-role" in self.properties.path

    async def _apply_to_account(  # noqa: C901
        self, aws_account: AWSAccount
    ) -> AccountChangeDetails:
        client = await aws_account.get_boto3_client("iam")
        account_role = self.apply_resource_dict(aws_account)
        role_name = account_role["RoleName"]
        account_change_details = AccountChangeDetails(
            account=str(aws_account),
            resource_id=role_name,
            resource_type=self.resource_type,
            new_value=dict(**account_role),
            proposed_changes=[],
            exceptions_seen=[],
        )
        if self._is_iambic_import_only():
            # Don't apply changes to aws-service-roles
            return account_change_details

        log_params = dict(
            resource_type=self.resource_type,
            resource_id=role_name,
            account=str(aws_account),
        )
        deleted = self.get_attribute_val_for_account(aws_account, "deleted", False)
        current_role = await get_role(
            role_name, client, include_policies=bool(not deleted)
        )
        if current_role:
            account_change_details.current_value = {**current_role}  # Create a new dict

            if ctx.command == Command.CONFIG_DISCOVERY:
                # Don't overwrite a resource during config discovery
                account_change_details.new_value = {}
                return account_change_details

        deleted = self.get_attribute_val_for_account(aws_account, "deleted", False)
        if isinstance(deleted, list):
            deleted = deleted[0].deleted

        if deleted:
            if current_role:
                account_change_details.new_value = None
                proposed_changes = [
                    ProposedChange(
                        change_type=ProposedChangeType.DELETE,
                        resource_id=role_name,
                        resource_type=self.resource_type,
                    )
                ]
                log_str = "Active resource found with deleted=false."
                if ctx.execute:
                    log_str = f"{log_str} Deleting resource..."
                log.debug(log_str, **log_params)

                if ctx.execute:
                    apply_awaitable = delete_iam_role(role_name, client, log_params)
                    proposed_changes = await plugin_apply_wrapper(
                        apply_awaitable, proposed_changes
                    )

                account_change_details.extend_changes(proposed_changes)

            return account_change_details

        role_exists = bool(current_role)
        inline_policies = account_role.pop("InlinePolicies", [])
        managed_policies = account_role.pop("ManagedPolicies", [])
        existing_inline_policies = current_role.pop("InlinePolicies", [])
        existing_managed_policies = current_role.pop("ManagedPolicies", [])
        tasks = []
        if role_exists:
            tasks.extend(
                [
                    apply_role_tags(
                        role_name,
                        client,
                        account_role["Tags"],
                        current_role.get("Tags", []),
                        log_params,
                    ),
                    update_assume_role_policy(
                        role_name,
                        client,
                        account_role.pop("AssumeRolePolicyDocument", {}),
                        current_role["AssumeRolePolicyDocument"],
                        log_params,
                    ),
                    apply_role_permission_boundary(
                        role_name,
                        client,
                        account_role.get("PermissionsBoundary", {}),
                        current_role.get("PermissionsBoundary", {}),
                        log_params,
                    ),
                ]
            )

            supported_update_keys = ["Description", "MaxSessionDuration"]
            update_resource_log_params = {**log_params}
            update_role_params = {}
            for k in supported_update_keys:
                if account_role.get(k) is not None and account_role.get(
                    k
                ) != current_role.get(k):
                    update_resource_log_params[k] = dict(
                        old_value=current_role.get(k), new_value=account_role.get(k)
                    )
                    update_role_params[k] = current_role.get(k)

            if update_role_params:
                log_str = "Out of date resource found."
                if ctx.execute:
                    log.debug(
                        f"{log_str} Updating resource...",
                        **update_resource_log_params,
                    )

                    async def update_role():
                        exceptions = []
                        try:
                            await boto_crud_call(
                                client.update_role,
                                RoleName=role_name,
                                **{
                                    k: account_role.get(k)
                                    for k in supported_update_keys
                                },
                            )
                        except Exception as e:
                            exceptions.append(str(e))
                        return [
                            ProposedChange(
                                change_type=ProposedChangeType.UPDATE,
                                resource_id=role_name,
                                resource_type=self.resource_type,
                                exceptions_seen=exceptions,
                            )
                        ]

                    tasks.append(update_role())
                else:
                    log.debug(log_str, **update_resource_log_params)
                    account_change_details.proposed_changes.append(
                        ProposedChange(
                            change_type=ProposedChangeType.UPDATE,
                            resource_id=role_name,
                            resource_type=self.resource_type,
                        )
                    )
        else:
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.CREATE,
                    resource_id=role_name,
                    resource_type=self.resource_type,
                )
            ]
            log_str = "New resource found in code."
            if not ctx.execute:
                # Exit now because apply functions won't work if resource doesn't exist
                log.debug(log_str, **log_params)
                account_change_details.extend_changes(proposed_changes)
                return account_change_details

            log.debug(f"{log_str} Creating resource...", **log_params)
            account_role["AssumeRolePolicyDocument"] = json.dumps(
                account_role["AssumeRolePolicyDocument"]
            )
            if account_role.get("PermissionsBoundary"):
                account_role["PermissionsBoundary"] = account_role[
                    "PermissionsBoundary"
                ]["PolicyArn"]

            apply_awaitable = boto_crud_call(client.create_role, **account_role)
            account_change_details.extend_changes(
                await plugin_apply_wrapper(apply_awaitable, proposed_changes)
            )

        tasks.extend(
            [
                apply_role_managed_policies(
                    role_name,
                    client,
                    managed_policies,
                    existing_managed_policies,
                    log_params,
                ),
                apply_role_inline_policies(
                    role_name,
                    client,
                    inline_policies,
                    existing_inline_policies,
                    log_params,
                ),
            ]
        )

        changes_made = await asyncio.gather(*tasks)
        if any(changes_made):
            account_change_details.extend_changes(
                list(chain.from_iterable(changes_made))
            )

        if ctx.execute and not account_change_details.exceptions_seen:
            log.debug(
                "Successfully finished execution on account for resource",
                changes_made=bool(account_change_details.proposed_changes),
                **log_params,
            )
        elif account_change_details.exceptions_seen:
            log.error(
                "Unable to finish execution on account for resource",
                exceptions_seen=[
                    cd.exceptions_seen for cd in account_change_details.exceptions_seen
                ],
                **log_params,
            )
        else:
            log.debug(
                "Successfully finished scanning for drift on account for resource",
                requires_changes=bool(account_change_details.proposed_changes),
                **log_params,
            )

        return account_change_details

    @classmethod
    def iambic_specific_knowledge(cls) -> set[str]:
        return {"access_rules", "file_path", "iambic_managed"}



from __future__ import annotations

import pathlib
import resource
from typing import TYPE_CHECKING

from iambic.core.logger import log
from iambic.core.utils import gather_templates

if TYPE_CHECKING:
    from iambic.config.dynamic_config import Config


async def resolve_config_template_path(repo_dir: str) -> pathlib.Path:
    config_template_file_path = await gather_templates(repo_dir, "Core::Config")
    if len(config_template_file_path) == 0:
        raise RuntimeError(
            f"Unable to discover IAMbic Configuration in {repo_dir}. "
            "Create a configuration with the `NOQ::Core::Config` template type. "
            "You can run `iambic setup` to generate a configuration."
        )
    if len(config_template_file_path) > 1:
        # IAMbic supports 1 configuration per repo
        raise RuntimeError(
            f"Too many NOQ::Core::Config templates discovered. Found ({len(config_template_file_path)}). "
            f"Expected 1. Files: {','.join([str(x) for x in config_template_file_path])}"
        )

    return pathlib.Path(config_template_file_path[0])


def check_and_update_resource_limit(config: Config):
    soft_limit, hard_limit = resource.getrlimit(resource.RLIMIT_NOFILE)
    minimum_ulimit = 64000
    if config.core:
        minimum_ulimit = config.core.minimum_ulimit
    if soft_limit < minimum_ulimit:
        try:
            resource.setrlimit(resource.RLIMIT_NOFILE, (minimum_ulimit, hard_limit))
        except PermissionError:
            log.warning(
                "Cannot increase resource limit: the process does not have permission.",
                current_ulimit=soft_limit,
                desired_ulimit=minimum_ulimit,
            )
        except Exception:
            log.warning(
                "Unable to increase resource limit: please manually update the soft limit.",
                current_ulimit=soft_limit,
                desired_ulimit=minimum_ulimit,
            )



from __future__ import annotations

import asyncio
import importlib
import itertools
import os
import subprocess
from collections import defaultdict
from enum import Enum
from pathlib import Path
from typing import List, Optional, Union
from uuid import uuid4

import ujson as json
from pydantic import BaseModel, Field
from pydantic import create_model as create_pydantic_model

import iambic.plugins.v0_1_0.github
from iambic.core.context import ctx
from iambic.core.exceptions import MultipleSecretsNotAcceptedException
from iambic.core.iambic_plugin import ProviderPlugin
from iambic.core.logger import log
from iambic.core.models import BaseTemplate, ExecutionMessage, TemplateChangeDetails
from iambic.core.utils import sort_dict, yaml
from iambic.plugins.v0_1_0 import PLUGIN_VERSION, aws, azure_ad, google_workspace, okta

CURRENT_IAMBIC_VERSION = "1"


class CoreConfig(BaseModel):
    minimum_ulimit: int = 64000


class PluginType(Enum):
    DIRECTORY_PATH = "DIRECTORY_PATH"
    # GIT = "GIT"


class ExtendsConfigKey(Enum):
    AWS_SECRETS_MANAGER = "AWS_SECRETS_MANAGER"
    LOCAL_FILE = "LOCAL_FILE"


class PluginDefinition(BaseModel):
    type: PluginType
    location: str = Field(
        description="The location of the plugin. "
        "For a DIRECTORY_PATH, this is the path to the plugin. "
        "For a GIT plugin, this is the git url."
    )
    version: str


def load_plugins(
    plugin_defs: list[PluginDefinition],
) -> List[tuple[ProviderPlugin, BaseModel]]:
    """
    Load all plugins from the given paths.
    This will search recursively for all files ending in iambic_plugin.py
    It will retrieve the variable IAMBIC_PLUGIN set in the iambic_plugin module.

    :param plugin_defs: A list of PluginDefinition instances to load.
    :return: A list of tuples containing the plugin and the plugin config model.
    """
    plugins = []
    for plugin_def in plugin_defs:
        if plugin_def.type == PluginType.DIRECTORY_PATH:
            for root, dirs, files in os.walk(plugin_def.location):
                for file in files:
                    if file == "iambic_plugin.py":
                        module_name, _ = os.path.splitext(file)
                        module_path = os.path.join(root, file)
                        module = importlib.machinery.SourceFileLoader(
                            module_name, module_path
                        ).load_module()
                        if iambic_plugin := getattr(module, "IAMBIC_PLUGIN"):
                            plugins.append(iambic_plugin)
                        else:
                            log.error(
                                "Could not find the IAMBIC_PLUGIN variable",
                                plugin_file_path=file,
                            )
        # TODO: Add git support here.
        #  Would work by trying to find the cloned repo in a defined dir

    log.debug("Loading discovered plugins", discovered_plugin_count=len(plugins))
    return plugins


class ExtendsConfig(BaseModel):
    key: ExtendsConfigKey
    value: str
    assume_role_arn: Optional[str]
    external_id: Optional[str]


class Config(BaseTemplate):
    template_type: str = "NOQ::Core::Config"
    version: str = Field(
        description="Do not change! The version of iambic this repo is compatible with.",
    )
    plugins: Optional[list[PluginDefinition]] = Field(
        default=[
            PluginDefinition(
                type=PluginType.DIRECTORY_PATH,
                location=aws.__path__[0],
                version=PLUGIN_VERSION,
            ),
            PluginDefinition(
                type=PluginType.DIRECTORY_PATH,
                location=google_workspace.__path__[0],
                version=PLUGIN_VERSION,
            ),
            PluginDefinition(
                type=PluginType.DIRECTORY_PATH,
                location=okta.__path__[0],
                version=PLUGIN_VERSION,
            ),
            PluginDefinition(
                type=PluginType.DIRECTORY_PATH,
                location=iambic.plugins.v0_1_0.github.__path__[0],
                version=PLUGIN_VERSION,
            ),
            PluginDefinition(
                type=PluginType.DIRECTORY_PATH,
                location=azure_ad.__path__[0],
                version=PLUGIN_VERSION,
            ),
        ],
        description="The plugins used by your IAMbic template repo.",
    )
    extends: List[ExtendsConfig] = []
    secrets: Optional[dict] = Field(
        description="Secrets should only be used in memory and never serialized out",
        exclude=True,
        default={},
    )
    plugin_instances: Optional[list[ProviderPlugin]] = Field(
        description="A list of the plugin instances parsed as part of the plugin paths.",
        exclude=True,
    )
    core: Optional[CoreConfig] = Field(
        CoreConfig(),
        description="Core configuration options for iambic.",
    )

    class Config:
        arbitrary_types_allowed = True

    def get_config_plugin(self, plugin: ProviderPlugin):
        return getattr(self, plugin.config_name)

    def set_config_plugin(self, plugin: ProviderPlugin, config: BaseModel):
        setattr(self, plugin.config_name, config)

    @property
    def configured_plugins(self):
        return [
            plugin for plugin in self.plugin_instances if self.get_config_plugin(plugin)
        ]

    async def set_config_secrets(self):
        if self.extends:
            if len(self.extends) > 1:
                raise MultipleSecretsNotAcceptedException()

            extend = self.extends[0]
            if extend.key == ExtendsConfigKey.LOCAL_FILE:
                dir_path = os.path.dirname(self.file_path)
                extend_path = os.path.join(dir_path, extend.value)
                with open(extend_path, "r") as ymlfile:
                    extend_config = yaml.load(ymlfile)
                for k, v in extend_config.items():
                    if not getattr(self, k, None):
                        setattr(self, k, v)
            else:
                decoded_secret_responses: list[dict] = await asyncio.gather(
                    *[
                        plugin.async_decode_secret_callable(
                            self.get_config_plugin(plugin), extend
                        )
                        for plugin in self.plugin_instances
                        if plugin.async_decode_secret_callable
                    ]
                )
                for decoded_secret in decoded_secret_responses:
                    if decoded_secret:
                        if "secrets" not in decoded_secret:
                            decoded_secret = dict(secrets=decoded_secret)
                        else:
                            for k, v in decoded_secret.items():
                                if k != "secrets":
                                    decoded_secret["secrets"][k] = v

                        for k, v in decoded_secret.get("secrets", {}).items():
                            if k in decoded_secret:
                                log.warning(
                                    "Secret key already exists. "
                                    "This means it was defined in multiple secrets"
                                    " or multiple times in the same secret.",
                                    key=k,
                                )
                            self.secrets.setdefault(k, v)

    async def run_import(
        self,
        exe_message: ExecutionMessage,
        output_dir: str,
    ):
        ctx.command = exe_message.parent_command
        # It's the responsibility of the provider to handle throttling.
        if exe_message.provider_type:
            plugin = [
                plugin
                for plugin in self.configured_plugins
                if plugin.config_name == exe_message.provider_type
            ][0]
            await plugin.async_import_callable(
                exe_message, self.get_config_plugin(plugin), output_dir
            )
        else:
            tasks = []
            for plugin in self.configured_plugins:
                task_message = exe_message.copy()
                task_message.provider_type = plugin.config_name
                tasks.append(
                    plugin.async_import_callable(
                        task_message, self.get_config_plugin(plugin), output_dir
                    )
                )

            await asyncio.gather(*tasks)

    async def run_apply(
        self, exe_message: ExecutionMessage, templates: list[BaseTemplate]
    ) -> list[TemplateChangeDetails]:
        # It's the responsibility of the provider to handle throttling.
        # Build a map of a plugin's template types to the plugin
        ctx.command = exe_message.parent_command
        template_provider_map = {}
        for plugin in self.plugin_instances:
            for template in plugin.templates:
                template_provider_map[
                    template.__fields__["template_type"].default
                ] = plugin.config_name

        # Create a map of the templates to apply to the template's plugin
        plugin_templates = defaultdict(list)
        for template in templates:
            provider = template_provider_map[template.template_type]
            plugin_templates[provider].append(template)

        tasks = []
        for plugin in self.plugin_instances:
            if (
                exe_message.provider_type
                and plugin.config_name != exe_message.provider_type
            ):
                continue

            if templates := plugin_templates.get(plugin.config_name):
                if plugin_config := self.get_config_plugin(plugin):
                    task_message = exe_message.copy()
                    task_message.provider_type = plugin.config_name
                    tasks.append(
                        plugin.async_apply_callable(
                            task_message, plugin_config, templates
                        )
                    )
                else:
                    log.warning(
                        "Templates discovered for a plugin not defined in the config file.",
                        missing_config=plugin.config_name,
                    )

        # Retrieve template changes across plugins and flatten responses
        template_changes = await asyncio.gather(*tasks)
        template_changes = list(itertools.chain.from_iterable(template_changes))

        if ctx.execute and template_changes:
            log.info("Finished applying changes.")
        elif not ctx.execute:
            log.info("Finished scanning for changes.")
        else:
            log.info("No changes found.")

        return template_changes

    async def run_detect_changes(
        self, repo_dir: str, run_import_as_fallback: bool = False
    ) -> Union[str, None]:
        change_str_list = await asyncio.gather(
            *[
                plugin.async_detect_changes_callable(
                    self.get_config_plugin(plugin), repo_dir
                )
                for plugin in self.configured_plugins
                if plugin.async_detect_changes_callable
            ]
        )
        if run_import_as_fallback:
            await asyncio.gather(
                *[
                    plugin.async_import_callable(
                        self.get_config_plugin(plugin), repo_dir
                    )
                    for plugin in self.configured_plugins
                    if not plugin.async_detect_changes_callable
                ]
            )

        if change_str_list := [
            change_str for change_str in change_str_list if change_str
        ]:
            return "\n".join(change_str_list)

    async def run_discover_upstream_config_changes(
        self, exe_message: ExecutionMessage, repo_dir: str
    ):
        ctx.command = exe_message.parent_command
        tasks = []
        for plugin in self.configured_plugins:
            if plugin.async_discover_upstream_config_changes_callable:
                task_message = exe_message.copy()
                task_message.provider_type = plugin.config_name
                tasks.append(
                    plugin.async_discover_upstream_config_changes_callable(
                        task_message, self.get_config_plugin(plugin), repo_dir
                    )
                )

        if tasks:
            log.info("Scanning for upstream changes to config attributes.")
            await asyncio.gather(*tasks)

        log.info("Finished scanning for upstream changes to config attributes.")
        self.write()

    async def configure_plugins(self):
        """
        Called to set plugin metadata that is generated at run-time.
        """

        # Sync to prevent issues updating the config
        for plugin in self.configured_plugins:
            if not plugin.requires_secret:
                await plugin.async_load_callable(self.get_config_plugin(plugin))

        await self.set_config_secrets()

        for plugin in self.plugin_instances:
            if plugin.requires_secret:
                if provider_config_dict := self.secrets.get(plugin.config_name):
                    try:
                        setattr(
                            self,
                            plugin.config_name,
                            plugin.provider_config(
                                **json.loads(json.dumps(provider_config_dict))
                            ),
                        )
                        await plugin.async_load_callable(self.get_config_plugin(plugin))
                    except Exception as err:
                        log.critical(
                            "Failed to configure plugin.",
                            err=err,
                            plugin=plugin.config_name,
                        )
                        raise

    def dict(
        self,
        *,
        include: Optional[
            Union["AbstractSetIntStr", "MappingIntStrAny"]  # noqa
        ] = None,
        exclude: Optional[
            Union["AbstractSetIntStr", "MappingIntStrAny"]  # noqa
        ] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = True,
        exclude_defaults: bool = False,
        exclude_none: bool = True,
    ) -> "DictStrAny":  # noqa
        required_exclude = {"secrets", "file_path", "plugin_instances"}

        if self.plugin_instances:
            for plugin in self.plugin_instances:
                if plugin.requires_secret:
                    required_exclude.add(plugin.config_name)

        if exclude:
            exclude.update(required_exclude)
        else:
            exclude = required_exclude

        return super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

    def write(self, exclude_none=True, exclude_unset=False, exclude_defaults=True):
        input_dict = self.dict(
            exclude_none=exclude_none,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
        )
        sorted_input_dict = sort_dict(
            input_dict,
            [
                "template_type",
                "version",
            ],
        )

        file_path = os.path.expanduser(self.file_path)
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w") as f:
            f.write(yaml.dump(sorted_input_dict))

        log.info("Config successfully written", config_location=file_path)


async def load_config(
    config_path: str,
    configure_plugins: bool = True,
    approved_plugins_only: bool = False,
) -> Config:
    """
    Load the configuration from the specified file path.

    Only use this function to retrieve the configuration object.
    Do not try to instantiate directly.
    Should be called as soon as possible in the application.
    This function is responsible for:
    - Loading the configuration file
    - Loading the plugins
    - Creating the dynamic configuration object using the loaded plugins
    - Setting runtime metadata for the plugins
    - Retrieving secrets which are used to set certain plugins
    - Setting list of templates on the TEMPLATE global variable

    Parameters:
    - config_path (str): The file path of the configuration file.

    Returns:
    - Config: The configuration object created from the specified file.
    """
    log.info("Loading config...")
    config_path = str(
        config_path
    )  # Ensure it's a string in case it's a Path for pydantic
    config_dict = yaml.load(open(config_path))
    base_config = Config(file_path=config_path, **config_dict)
    return await process_config(
        base_config, config_path, config_dict, configure_plugins, approved_plugins_only
    )


async def process_config(
    base_config: Config,
    config_path,
    config_dict,
    configure_plugins: bool = True,
    approved_plugins_only: bool = False,
) -> Config:
    from iambic.config.templates import TEMPLATES

    if approved_plugins_only:
        default_plugins = [
            plugin.location for plugin in Config.__fields__["plugins"].default
        ]
        base_config.plugins = [
            plugin
            for plugin in base_config.plugins
            if plugin.location in default_plugins
        ]

    all_plugins = load_plugins(base_config.plugins)
    config_fields = {}
    for plugin in all_plugins:
        config_fields[plugin.config_name] = (plugin.provider_config, None)

    dynamic_config = create_pydantic_model(
        f"DynamicConfig-{uuid4()}", __base__=Config, **config_fields
    )
    config = dynamic_config(
        plugin_instances=all_plugins, file_path=config_path, **config_dict
    )

    if configure_plugins:
        log.info("Setting config metadata...")
        await config.configure_plugins()
        log.info("Plugins loaded successfully...")

    TEMPLATES.set_templates(
        list(
            itertools.chain.from_iterable(
                [plugin.templates for plugin in config.plugin_instances]
            )
        )
    )
    return config


async def init_plugins(config_path: str):
    """
    Download plugins and install plugin dependencies
    """
    config_path = str(
        config_path
    )  # Ensure it's a string in case it's a Path for pydantic
    config_dict = yaml.load(open(config_path))
    config = Config(file_path=config_path, **config_dict)
    errors = defaultdict(dict)

    for plugin in config.plugins:
        if plugin.type == PluginType.DIRECTORY_PATH:
            plugin_path = plugin.location
        # elif plugin.type == PluginType.GIT:
        # Clone repo, set plugin_path

        if "aws" in plugin_path:
            continue

        required_files = ["iambic_plugin.py", "pyproject.toml"]
        for required_file in required_files:
            file_path = Path(os.path.join(plugin_path, required_file))
            if not file_path.is_file():
                errors[plugin.location].setdefault("missing_required_file", []).append(
                    str(file_path)
                )

        if plugin.location in errors:
            continue

        try:
            log.info("Installing plugin dependencies", plugin=plugin.location)
            subprocess.run(
                [
                    "poetry",
                    "install",
                ],
                cwd=plugin_path,
                check=True,
                capture_output=True,
                text=True,
            )
        except subprocess.CalledProcessError as err:
            errors[plugin.location]["install_failed"] = dict(error=err.stderr)
        except Exception as err:
            errors[plugin.location]["install_failed"] = dict(error=repr(err))

    if errors:
        log.error("Failed to initialize all plugins", **errors)



from __future__ import annotations

import unittest.mock

import pytest
from mock import AsyncMock
from pydantic import SecretStr

from iambic.core.iambic_enum import IambicManaged
from iambic.core.models import Command, ExecutionMessage
from iambic.plugins.v0_1_0.google_workspace.handlers import import_google_resources
from iambic.plugins.v0_1_0.google_workspace.iambic_plugin import (
    GoogleProject,
    GoogleSubject,
    GoogleWorkspaceConfig,
)


@pytest.mark.asyncio
async def test_import_google_resources():
    # Mock ExecutionMessage
    exe_message = ExecutionMessage(
        execution_id="11928f24-11f4-4629-a38f-9c24c5551e89",
        command=Command.IMPORT,
        parent_command=Command.IMPORT,
        provider_type=None,
        provider_id=None,
        template_type=None,
        template_id=None,
        metadata=None,
        templates=None,
    )

    # Mock GoogleWorkspaceConfig
    config = GoogleWorkspaceConfig(
        workspaces=[
            GoogleProject(
                project_id="example-ga",
                project_name=None,
                subjects=[
                    GoogleSubject(
                        domain="example.com",
                        service_account="google-groups-test@example-ga.iam.gserviceaccount.com",
                    )
                ],
                type="service_account",
                private_key_id="12345",
                private_key=SecretStr("**********"),
                client_email="google-groups-test@example-ga.iam.gserviceaccount.com",
                client_id="12345",
                auth_uri="https://accounts.google.com/o/oauth2/auth",
                token_uri="https://oauth2.googleapis.com/token",
                auth_provider_x509_cert_url="https://www.googleapis.com/oauth2/v1/certs",
                client_x509_cert_url="https://www.googleapis.com/robot/v1/metadata/x509/google-groups-test%40example-ga.iam.gserviceaccount.com",
                variables=[],
                iambic_managed=IambicManaged.UNDEFINED,
            )
        ]
    )
    base_output_dir = "../example-templates-local"
    messages = None
    remote_worker = None

    # Mock the collect_project_groups and generate_group_templates functions
    with unittest.mock.patch(
        "iambic.plugins.v0_1_0.google_workspace.handlers.collect_project_groups",
        new_callable=AsyncMock,
    ) as mock_collect_project_groups, unittest.mock.patch(
        "iambic.plugins.v0_1_0.google_workspace.handlers.generate_group_templates",
        new_callable=AsyncMock,
    ) as mock_generate_group_templates:
        # Call the import_google_resources function
        await import_google_resources(
            exe_message, config, base_output_dir, messages, remote_worker
        )

        # Assert that the mocked functions were called with the expected arguments
        mock_collect_project_groups.assert_called()
        mock_generate_group_templates.assert_called_with(
            exe_message, config, base_output_dir
        )



from __future__ import annotations

import asyncio
from itertools import chain
from typing import Any, Callable, Optional, Union

from pydantic import Field, validator

from iambic.core.context import ctx
from iambic.core.iambic_enum import Command
from iambic.core.logger import log
from iambic.core.models import (
    AccountChangeDetails,
    BaseModel,
    ExpiryModel,
    ProposedChange,
    ProposedChangeType,
)
from iambic.core.utils import plugin_apply_wrapper, remove_expired_resources
from iambic.plugins.v0_1_0.aws.iam.models import Path, PermissionBoundary
from iambic.plugins.v0_1_0.aws.iam.policy.models import ManagedPolicyRef, PolicyDocument
from iambic.plugins.v0_1_0.aws.iam.user.utils import (
    apply_user_groups,
    apply_user_inline_policies,
    apply_user_managed_policies,
    apply_user_permission_boundary,
    apply_user_tags,
    delete_iam_user,
    get_user,
)
from iambic.plugins.v0_1_0.aws.models import AccessModel, AWSAccount, AWSTemplate, Tag
from iambic.plugins.v0_1_0.aws.utils import boto_crud_call

AWS_IAM_USER_TEMPLATE_TYPE = "NOQ::AWS::IAM::User"


class Group(ExpiryModel, AccessModel):
    group_name: str

    # All excluded fields are populated by the API and not required in the template
    # We are not tracking those, but we do allow them to be imported in order to
    # pass validation.
    arn: Optional[str] = Field("", description="ARN of the group", exclude=True)
    create_date: Optional[str] = Field(
        "", description="Date the group was created", exclude=True
    )
    group_id: Optional[str] = Field("", description="ID of the group", exclude=True)
    path: Optional[str] = Field("", description="Path of the group", exclude=True)
    extra: Any = Field(None, description="Extra attributes to store", exclude=True)

    @property
    def resource_type(self):
        return "aws:iam:group"

    @property
    def resource_id(self):
        return self.group_name


class UserProperties(BaseModel):
    user_name: str = Field(
        description="Name of the user",
    )
    path: Optional[Union[str, list[Path]]] = "/"
    permissions_boundary: Optional[
        Union[None, PermissionBoundary, list[PermissionBoundary]]
    ] = None
    tags: Optional[list[Tag]] = Field(
        [],
        description="List of tags attached to the user",
    )
    groups: Optional[list[Group]] = Field(
        [],
        description="List of groups the user is a member of",
    )
    managed_policies: Optional[list[ManagedPolicyRef]] = Field(
        [],
        description="Managed policy arns attached to the user",
    )
    inline_policies: Optional[list[PolicyDocument]] = Field(
        [],
        description="List of the user's inline policies",
    )

    @property
    def resource_type(self):
        return "aws:iam:user"

    @property
    def resource_id(self):
        return self.user_name

    @classmethod
    def sort_func(cls, attribute_name: str) -> Callable:
        def _sort_func(obj):
            return f"{getattr(obj, attribute_name)}!{obj.access_model_sort_weight()}"

        return _sort_func

    @validator("tags")
    def sort_tags(cls, v: list[Tag]):
        sorted_v = sorted(v, key=cls.sort_func("key"))
        return sorted_v

    @validator("groups")
    def sort_groups(cls, v: list[Group]):
        sorted_v = sorted(v, key=cls.sort_func("group_name"))
        return sorted_v

    @validator("managed_policies")
    def sort_managed_policy_refs(cls, v: list[ManagedPolicyRef]):
        sorted_v = sorted(v, key=cls.sort_func("policy_arn"))
        return sorted_v

    @validator("inline_policies")
    def sort_inline_policies(cls, v: list[PolicyDocument]):
        sorted_v = sorted(v, key=cls.sort_func("policy_name"))
        return sorted_v

    @validator("path")
    def sort_path(cls, v: list[Path]):
        if not isinstance(v, list):
            return v
        sorted_v = sorted(v, key=lambda d: d.access_model_sort_weight())
        return sorted_v

    @validator("permissions_boundary")
    def sort_permissions_boundary(cls, v: list[PermissionBoundary]):
        if not isinstance(v, list):
            return v
        sorted_v = sorted(v, key=lambda d: d.access_model_sort_weight())
        return sorted_v


class AwsIamUserTemplate(AWSTemplate, AccessModel):
    template_type = AWS_IAM_USER_TEMPLATE_TYPE
    properties: UserProperties = Field(
        description="Properties of the user",
    )

    def _apply_resource_dict(self, aws_account: AWSAccount = None) -> dict:
        response = super(AwsIamUserTemplate, self)._apply_resource_dict(aws_account)
        if "Tags" not in response:
            response["Tags"] = []

        if permissions_boundary := response.pop("PermissionsBoundary", []):
            if isinstance(permissions_boundary, list):
                permissions_boundary = permissions_boundary[0]
            response["PermissionsBoundary"] = permissions_boundary

        if isinstance(response.get("Description"), list):
            response["Description"] = response["Description"][0]["Description"]

        return response

    async def _apply_to_account(  # noqa: C901
        self, aws_account: AWSAccount
    ) -> AccountChangeDetails:
        client = await aws_account.get_boto3_client("iam")
        self = await remove_expired_resources(
            self, self.resource_type, self.resource_id
        )
        account_user = self.apply_resource_dict(aws_account)

        user_name = account_user["UserName"]
        account_change_details = AccountChangeDetails(
            account=str(aws_account),
            resource_id=user_name,
            resource_type=self.resource_type,
            new_value=dict(**account_user),
            proposed_changes=[],
            exceptions_seen=[],
        )
        log_params = dict(
            resource_type=self.resource_type,
            resource_id=user_name,
            account=str(aws_account),
        )
        deleted = self.get_attribute_val_for_account(aws_account, "deleted", False)
        current_user = await get_user(
            user_name, client, include_policies=bool(not deleted)
        )
        if current_user:
            account_change_details.current_value = {**current_user}  # Create a new dict

            if ctx.command == Command.CONFIG_DISCOVERY:
                # Don't overwrite a resource during config discovery
                account_change_details.new_value = {}
                return account_change_details

        deleted = self.get_attribute_val_for_account(aws_account, "deleted", False)
        if isinstance(deleted, list):
            deleted = deleted[0].deleted

        if deleted:
            if current_user:
                account_change_details.new_value = None
                log_str = "Active resource found with deleted=false."
                if ctx.execute:
                    log_str = f"{log_str} Deleting resource..."
                log.debug(log_str, **log_params)

                proposed_changes = [
                    ProposedChange(
                        change_type=ProposedChangeType.DELETE,
                        resource_id=user_name,
                        resource_type=self.resource_type,
                    )
                ]
                if ctx.execute:
                    apply_awaitable = delete_iam_user(user_name, client, log_params)
                    proposed_changes = await plugin_apply_wrapper(
                        apply_awaitable, proposed_changes
                    )

                account_change_details.extend_changes(proposed_changes)

            return account_change_details

        user_exists = bool(current_user)
        inline_policies = account_user.pop("InlinePolicies", [])
        managed_policies = account_user.pop("ManagedPolicies", [])
        groups = account_user.pop("Groups", [])
        existing_inline_policies = current_user.pop("InlinePolicies", [])
        existing_managed_policies = current_user.pop("ManagedPolicies", [])
        existing_groups = current_user.pop("Groups", [])
        tasks = []

        if user_exists:
            tasks.extend(
                [
                    apply_user_tags(
                        user_name,
                        client,
                        account_user["Tags"],
                        current_user.get("Tags", []),
                        log_params,
                    ),
                    apply_user_permission_boundary(
                        user_name,
                        client,
                        account_user.get("PermissionsBoundary", {}),
                        current_user.get("PermissionsBoundary", {}),
                        log_params,
                    ),
                ]
            )

            supported_update_keys = ["Path", "UserName"]
            update_resource_log_params = {**log_params}
            update_user_params = {}
            for k in supported_update_keys:
                if account_user.get(k) is not None and account_user.get(
                    k
                ) != current_user.get(k):
                    update_resource_log_params[k] = dict(
                        old_value=current_user.get(k), new_value=account_user.get(k)
                    )
                    update_user_params[f"New{k}"] = account_user.get(k)
            if update_user_params:
                log_str = "Out of date resource found."
                if ctx.execute:
                    log.info(
                        f"{log_str} Updating resource...",
                        **update_resource_log_params,
                    )

                    async def update_user():
                        exceptions = []
                        try:
                            await boto_crud_call(
                                client.update_user,
                                UserName=user_name,
                                **{
                                    k: update_user_params.get(k)
                                    for k in update_user_params
                                },
                            )
                        except Exception as e:
                            exceptions.append(str(e))
                        return [
                            ProposedChange(
                                change_type=ProposedChangeType.UPDATE,
                                resource_id=user_name,
                                resource_type=self.resource_type,
                                exceptions_seen=exceptions,
                            )
                        ]

                    tasks.append(update_user())
                else:
                    log.info(log_str, **update_resource_log_params)
                    account_change_details.proposed_changes.append(
                        ProposedChange(
                            change_type=ProposedChangeType.UPDATE,
                            resource_id=user_name,
                            resource_type=self.resource_type,
                        )
                    )
        else:
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.CREATE,
                    resource_id=user_name,
                    resource_type=self.resource_type,
                )
            ]
            log_str = "New resource found in code."
            if not ctx.execute:
                # Exit now because apply functions won't work if resource doesn't exist
                log.debug(log_str, **log_params)
                account_change_details.extend_changes(proposed_changes)
                return account_change_details

            log_str = f"{log_str} Creating resource..."
            log.info(log_str, **log_params)
            if account_user.get("PermissionsBoundary"):
                account_user["PermissionsBoundary"] = account_user[
                    "PermissionsBoundary"
                ]["PolicyArn"]

            account_change_details.extend_changes(
                await plugin_apply_wrapper(
                    boto_crud_call(client.create_user, **account_user),
                    proposed_changes,
                )
            )

        tasks.extend(
            [
                apply_user_managed_policies(
                    user_name,
                    client,
                    managed_policies,
                    existing_managed_policies,
                    log_params,
                ),
                apply_user_inline_policies(
                    user_name,
                    client,
                    inline_policies,
                    existing_inline_policies,
                    log_params,
                ),
                apply_user_groups(
                    user_name,
                    client,
                    groups,
                    existing_groups,
                    log_params,
                ),
            ]
        )

        changes_made = await asyncio.gather(*tasks)
        if any(changes_made):
            account_change_details.extend_changes(
                list(chain.from_iterable(changes_made))
            )

        if ctx.execute and not account_change_details.exceptions_seen:
            log.debug(
                "Successfully finished execution on account for resource",
                changes_made=bool(account_change_details.proposed_changes),
                **log_params,
            )
        elif account_change_details.exceptions_seen:
            log.error(
                "Unable to finish execution on account for resource",
                exceptions_seen=[
                    cd.exceptions_seen for cd in account_change_details.exceptions_seen
                ],
                **log_params,
            )
        else:
            log.debug(
                "Successfully finished scanning for drift on account for resource",
                requires_changes=bool(account_change_details.proposed_changes),
                **log_params,
            )

        return account_change_details
